import os
import os.path
import io
import gettext
import codecs

from django.conf import settings

from pytigon_lib.schindent.py_to_js import compile
from pytigon_lib.schindent.indent_tools import convert_js
from pytigon_lib.schtools.tools import norm_indent
from pytigon_lib.schtools.main_paths import get_prj_name


try:
    import markdown

    def convert_md(stream_in, stream_out):
        """
        Convert Markdown code to HTML and write to a stream.

        :param stream_in: Input stream with Markdown code
        :param stream_out: Output stream with HTML code
        :return: True if conversion was successful, False otherwise
        """
        if stream_in and stream_out:
            buf = markdown.markdown(
                norm_indent(stream_in.getvalue()),
                extensions=[
                    "abbr",
                    "attr_list",
                    "def_list",
                    "fenced_code",
                    "footnotes",
                    "md_in_html",
                    "tables",
                    "admonition",
                    "codehilite",
                ],
            )
            stream_out.write(buf)
            return True
        return False

except:
    pass

PY_TO_JS = None


def list_with_next_generator(l):
    old = l[0]
    for pos in l[1:]:
        yield (old, pos)
        old = pos
    yield (l[-1], None)


def translate(s):
    return s


def iter_lines(f, f_name, lang):
    in_table = 0
    base_path = ""
    locale_path = ""

    tab_translate = []

    if lang != "en":
        base_path = os.path.join(settings.PRJ_PATH, get_prj_name())
        locale_path = os.path.join(base_path, "locale")
        try:
            try:
                t = gettext.translation(
                    "django",
                    locale_path,
                    languages=[
                        lang,
                    ],
                )
                t.install()
            except:
                t = None
            try:
                with open(
                    os.path.join(base_path, "translate.py"), "rt", encoding="utf-8"
                ) as p:
                    for line in p.readlines():
                        fr = line.split('_("')
                        if len(fr) > 1:
                            fr = fr[1].split('")')
                            if len(fr) == 2:
                                tab_translate.append(fr[0])
            except:
                tab_translate = []

            def trans(word):
                if len(word) < 2:
                    return word
                if word[0] == word[-1] == '"' or word[0] == word[-1] == "'":
                    if word[0] == "'":
                        strtest = "'"
                    else:
                        strtest = '"'
                    word2 = word[1:-1]
                else:
                    strtest = None
                    word2 = word

                if not word2 in tab_translate:
                    tab_translate.append(word2)
                if t:
                    ret = t.gettext(word2)
                    if strtest != None:
                        return strtest + ret + strtest
                    else:
                        return ret
                else:
                    return translate(word)

            gt = trans
        except:
            gt = translate
    else:
        gt = translate

    for line in f:
        # if len(line.lstrip()) == 0:
        #    continue
        if line.lstrip().startswith("_") and not line.lstrip().startswith("_("):
            nr = line.find("_")
            line2 = " " * nr + "." + gt(line.strip()[1:])
        else:
            if "_(" in line and not "__(" in line:
                out = []
                if line.lstrip().startswith("_("):
                    fr0 = line.split("_(")
                    fr = (fr0[0] + ".", fr0[1])
                else:
                    fr = line.split("_(")
                out.append(fr[0])
                for pos in fr[1:]:
                    id2 = pos.find(")")
                    if id2 >= 0:
                        out.append(gt(pos[:id2]))
                        out.append(pos[id2 + 1 :])
                    else:
                        out.append(gt(pos))
                line2 = "".join(out)
            else:
                line2 = line

        line3 = line2.strip()
        if len(line3) > 0 and (line3[0] == "[" or line3[-1] == "]" or "|" in line3):
            if line3[0] == "[" and (
                line3[-1] == "|" or (line3[-1] == "]" and "|" in line3)
            ):
                if line3[1] == "[":
                    in_table = 2
                else:
                    in_table = 1
            if in_table == 1:
                line2 = (
                    line2.replace("[", "<tr><td>")
                    .replace("]", "</td></tr>")
                    .replace(" |", " </td><td>")
                )
            if in_table == 2:
                line2 = (
                    line2.replace("[[", "<tr><th>")
                    .replace("]]", "</th></tr>")
                    .replace(" |", " </th><th>")
                )
            if line3[-1] == "]":
                in_table = False
        yield line2
    yield "."
    if len(tab_translate) > 0:
        if "site-packages" not in base_path:
            try:
                with open(
                    os.path.join(base_path, "translate.py"), "wt", encoding="utf-8"
                ) as p:
                    for word in tab_translate:
                        p.write('_("' + word + '")\n')
            except:
                import sys, traceback

                print(sys.exc_info())
                traceback.print_exc()


class ConwertToHtml:
    def __init__(
        self,
        file_name,
        simple_close_elem,
        auto_close_elem,
        no_auto_close_elem,
        input_str=None,
        lang="en",
        output_processors=None,
    ):
        self.file_name = file_name
        self.input_str = input_str
        self.code = []
        self.bufor = []
        self.output = []
        self.no_conwert = False
        self.simple_close_elem = simple_close_elem
        self.auto_close_elem = auto_close_elem
        self.no_auto_close_elem = no_auto_close_elem
        self.lang = lang
        self.output_processors = output_processors

    def _output_buf(self, nr):
        for pos in reversed(self.bufor):
            if pos[0] >= nr:
                self.output.append([pos[0], pos[1], pos[2]])
                self.bufor.remove(pos)
            else:
                break

    def _space_count(self, buf):
        i = 0
        for z in buf:
            if z == " ":
                i = i + 1
            else:
                break
        return i

    def _get_elem(self, elem):
        elem2 = elem.lstrip()
        id = elem2.find(" ")
        if id > 0:
            return elem2[:id]
        return elem2

    def _transform_elem(self, elem):
        id = elem.find(" ")
        if id > 0:
            elem0 = elem[:id]
            elem1 = ""
            tmp = elem[id + 1 :].split(",,,")
            for pos in tmp:
                id2 = pos.find("=")
                if id2 > 0:
                    elem1 = elem1 + " " + pos[:id2] + '="' + pos[id2 + 1 :] + '"'
                else:
                    elem1 = elem1 + " " + pos
            return elem0 + elem1
        else:
            return elem

    def _pre_process_line(self, line):
        n = self._space_count(line)
        line2 = line[n:]
        if line2.rstrip() == "":
            return [None]
        if not (
            (line2[0] >= "a" and line2[0] <= "z")
            or (line2[0] >= "A" and line2[0] <= "Z")
            or line2[0] == "%"
        ):
            if line2[0] == ".":
                return [(n, None, line2[1:], 0)]
            else:
                return [(n, None, line2, 0)]
        nr = line2.find("...")
        if nr >= 0:
            code = line2[:nr]
            html = line2[nr + 3 :]
            if code == "":
                code = None
        else:
            code = line2
            html = None
        if code:
            if code[0] != "%":
                code2 = code.split(":::")
                if len(code2) > 1:
                    out = []
                    i = n
                    for pos in code2[:-1]:
                        out.append(
                            (i, self._transform_elem(pos), None, 3 if i == n else 1)
                        )
                        i = i + 1
                    out.append((i, self._transform_elem(code2[-1]), html, 4))
                    return out
                else:
                    code = self._transform_elem(code)
        return [(n, code, html, 0)]

    def _status_close(self, status, line, next_line):
        if status == 0:
            return 0
        if status == 1:
            return 1
        if status == 2:
            return 0
        if status == 3:
            return 2
        if status == 4:
            if next_line[0] > line[0]:
                return 3
            else:
                return 1
        return status

    def transform_line(self, line, next_line):
        if not (line[1] == None and line[2] == None):
            self._output_buf(line[0])

        if line[1]:
            if line[1][0] == "%":
                if line[1][1] == "%":
                    if next_line[0] <= line[0] and not (
                        next_line[1] == None and next_line[2] == None
                    ):
                        self.output.append(
                            [
                                line[0],
                                "{% block "
                                + (line[1])[2:].lstrip()
                                + " %}"
                                + (line[2] if line[2] else "")
                                + "{% endblock %}",
                                line[3],
                            ]
                        )
                    else:
                        self.output.append(
                            [
                                line[0],
                                "{% block " + (line[1])[2:].lstrip() + " %}",
                                line[3],
                            ]
                        )
                        if line[2]:
                            self.output.append([line[0], line[2], line[3]])
                        self.bufor.append([line[0], "{% endblock %}", line[3]])
                else:
                    auto_end = False
                    tag = (line[1])[1:].split()[0].strip()
                    full_tag = (line[1])[1:].strip()
                    if full_tag.endswith(":"):
                        auto_end = True
                        tag = tag.replace(":", "")
                        full_tag = full_tag[:-1]
                        if tag in self.no_auto_close_elem:
                            auto_end = False

                    self.output.append([line[0], "{% " + full_tag + " %}", line[3]])

                    if auto_end or tag in self.auto_close_elem or "_ext" in tag:
                        self.bufor.append([line[0], "{% end" + tag + " %}", line[3]])
                    if line[2]:
                        self.output.append([line[0], line[2], line[3]])
            else:
                if next_line[0] <= line[0] and not (
                    next_line[1] == None and next_line[2] == None
                ):
                    if line[2] or not self._get_elem(line[1]) in self.simple_close_elem:
                        s = line[2] if line[2] else ""
                        self.output.append(
                            [
                                line[0],
                                "<"
                                + line[1]
                                + ">"
                                + s
                                + "</"
                                + self._get_elem(line[1])
                                + ">",
                                self._status_close(line[3], line, next_line),
                            ]
                        )
                    else:
                        self.output.append([line[0], "<" + line[1] + " />", line[3]])
                else:
                    self.output.append([line[0], "<" + line[1] + ">", line[3]])
                    if line[2]:
                        self.output.append([line[0], line[2], line[3]])
                    self.bufor.append(
                        [
                            line[0],
                            "</" + self._get_elem(line[1]) + ">",
                            self._status_close(line[3], line, next_line),
                        ]
                    )
        else:
            self.output.append([line[0], line[2], line[3]])

    def transform(self):
        old_line = None
        # breakpoint()
        for line in self.code:
            if old_line == None:
                old_line = line
            else:
                self.transform_line(old_line, line)
                old_line = line
        if old_line:
            self.transform_line(old_line, (0, None, None))
        self._output_buf(-1)

    def process(self):
        if self.file_name:
            file = codecs.open(self.file_name, "r", encoding="utf-8")
            x = file.readline()
            file.seek(0, 0)
            if x.startswith("@@@"):
                fname = x[3:].strip()
                fpath = os.path.join(os.path.dirname(self.file_name), fname) + ".ihtml"
                file2 = codecs.open(fpath, "r", encoding="utf-8")
                content2 = file.read()[len(x) :]
                content = file2.read().replace("@@@", content2)
                file2.close()
                file.close()
                file = io.StringIO(content)
        else:
            file = io.StringIO(self.input_str)
        old_pos = 0
        buf = None
        buf0 = ""
        test = 0
        cont = False
        indent_pos = 0
        for _line in iter_lines(file, self.file_name, self.lang):
            line = _line.replace("\n", "").replace("\r", "").replace("\t", "        ")
            if line.replace(" ", "") == "%else" or line.replace(" ", "") == "%else:":
                line = " " + line
            if "^^^" in line:
                self.no_conwert = True
                file.close()
                return

            if test:
                if (
                    test > 1
                    and len(line.strip()) > 0
                    and self._space_count(line) <= indent_pos
                ):
                    cont = True
                if cont or "<<<" in line:
                    if test == 1:
                        l = (
                            line.replace("\n", "")
                            .replace("\r", "")
                            .replace("\t", "        ")
                            .replace("<<<", "")
                            .rstrip()
                        )
                        buf.write(l)
                        x = self._pre_process_line(buf0 + buf.getvalue())
                        for pos in x:
                            if pos:
                                self.code.append((pos[0] * 4, pos[1], pos[2], pos[3]))
                                old_pos = pos[0]
                            else:
                                self.code.append((old_pos * 4, None, None, 1))
                        buf = None
                        test = 0
                        continue
                    if test > 1:
                        if not cont:
                            l = (
                                line.replace("\n", "")
                                .replace("\r", "")
                                .replace("\t", "        ")
                                .replace("<<<", "")
                                .rstrip()
                            )
                            buf.write(l)
                        if test == 2:
                            buf2 = io.StringIO()
                            convert_js(buf, buf2)
                            x = self._pre_process_line(buf0 + buf2.getvalue())
                        elif test == 3:
                            x = self._pre_process_line(
                                buf0.replace("pscript", "script language=python")
                                + buf.getvalue()
                            )
                            test = 0
                        elif test == 4:
                            v = buf.getvalue()
                            codejs = py_to_js(v, None)
                            x = self._pre_process_line(
                                buf0.replace("pscript", "script")
                                .replace(" language=python", "")
                                .replace("py2javascript", "javascript")
                                + codejs
                            )
                        elif test == 5:
                            v = buf.getvalue()
                            codejs = py_to_js(v, None)
                            x = self._pre_process_line(
                                buf0.replace("pscript", "script").replace(
                                    " language=python", ""
                                )
                                + codejs
                            )
                        elif test == 6:
                            buf2 = io.StringIO()
                            convert_md(buf, buf2)
                            x = self._pre_process_line(buf0 + buf2.getvalue())
                        else:
                            x = self._pre_process_line(buf0 + buf.getvalue())
                        for pos in x:
                            if pos:
                                self.code.append((pos[0] * 4, pos[1], pos[2], pos[3]))
                                old_pos = pos[0]
                            else:
                                self.code.append((old_pos * 4, None, None, 1))
                        buf = None
                        buf2 = None
                        test = 0
                else:
                    buf.write(
                        line.replace("\n", "")
                        .replace("\r", "")
                        .replace("\t", "        ")
                        .rstrip()
                        + "\n"
                    )
                    continue
            if cont or not test:
                cont = False
                if ">>>" in line:
                    pos = line.find(">>>")
                    if len(line[:pos].strip()) > 0:
                        buf0 = line[: pos + 3].replace(">>>", "...|||")
                    else:
                        buf0 = line[: pos + 3].replace(">>>", ".|||")
                    buf = io.StringIO()
                    buf.write(
                        line[pos + 3 :]
                        .replace("\n", "")
                        .replace("\r", "")
                        .replace("\t", "        ")
                        .rstrip()
                    )
                    test = 1
                elif "{:}" in line:
                    indent_pos = self._space_count(line)
                    pos = line.find("{:}")
                    if len(line[:pos].strip()) > 0:
                        buf0 = line[: pos + 4].replace("{:}", "...|||")
                    else:
                        buf0 = line[: pos + 4].replace("{:}", ".|||")
                    buf = io.StringIO()
                    buf.write(
                        line[pos + 4 :]
                        .replace("\n", "")
                        .replace("\r", "")
                        .replace("\t", "        ")
                        .rstrip()
                    )
                    test = 2
                elif "===>" in line:
                    indent_pos = self._space_count(line)
                    pos = line.find("===>")
                    if len(line[:pos].strip()) > 0:
                        buf0 = line[: pos + 4].replace("===>", "...|||")
                    else:
                        buf0 = line[: pos + 4].replace("===>", ".|||")
                    buf = io.StringIO()
                    buf.write(
                        line[pos + 4 :]
                        .replace("\n", "")
                        .replace("\r", "")
                        .replace("\t", "        ")
                        .rstrip()
                    )
                    test = 3
                elif "script language=python" in line:
                    indent_pos = self._space_count(line)
                    pos = line.find("script language=python")
                    buf0 = line + "...|||"
                    buf = io.StringIO()
                    buf.write(
                        line[pos + 22 :]
                        .replace("\n", "")
                        .replace("\r", "")
                        .replace("\t", "        ")
                        .rstrip()
                    )
                    test = 3
                elif "pscript" in line:
                    indent_pos = self._space_count(line)
                    buf0 = line + "...|||"
                    buf = io.StringIO()
                    test = 3
                elif (
                    line.strip().startswith("script")
                    and "language=py2javascript" in line
                ):
                    indent_pos = self._space_count(line)
                    buf0 = line + "...|||"
                    buf = io.StringIO()
                    test = 4
                elif line.strip().replace(" ", "").startswith("%component"):
                    indent_pos = self._space_count(line)
                    pos = line.rfind(":")
                    buf0 = line + "...|||"
                    test = 5
                    buf = io.StringIO()
                elif "###>" in line:
                    indent_pos = self._space_count(line)
                    pos = line.find("###>")
                    if len(line[:pos].strip()) > 0:
                        buf0 = line[: pos + 4].replace("###>", "...|||")
                    else:
                        buf0 = line[: pos + 4].replace("###>", ".|||")
                    buf = io.StringIO()
                    buf.write(
                        line[pos + 4 :]
                        .replace("\n", "")
                        .replace("\r", "")
                        .replace("\t", "        ")
                        .rstrip()
                    )
                    test = 6
                else:
                    l = (
                        line.replace("\n", "")
                        .replace("\r", "")
                        .replace("\t", "        ")
                        .rstrip()
                    )
                    x = self._pre_process_line(l)
                    for pos in x:
                        if pos:
                            self.code.append((pos[0] * 4, pos[1], pos[2], pos[3]))
                            old_pos = pos[0]
                        else:
                            self.code.append((old_pos * 4, None, None, 1))
        file.close()
        self.code.append((0, None, None, 1))
        self.transform()

    def to_str(self, beauty=True):
        if self.no_conwert:
            if self.file_name:
                file = codecs.open(self.file_name, "r", encoding="utf-8")
                output = file.read().replace("^^^", "")
                file.close()
            else:
                output = self.input_str.replace("^^^", "")
            output = output.replace("\r", "").replace("\\\n", "")
            return output
        else:
            output = ""
            if beauty:
                if len(self.output) > 0:
                    for line, nextline in list_with_next_generator(self.output):
                        if line[0] >= 0 and (line[2] == 0 or line[2] == 3):
                            output = output + " " * int(line[0] / 2)
                        if line[1]:
                            output = output + line[1]
                        if line[2] == 0 or line[2] == 2:
                            output = output + "\n"
                        if line[2] == 4 and nextline and nextline[0] > line[0]:
                            output = output + "\n"
                ret = output.replace("|||", "\n")
            else:
                if len(self.output) > 0:
                    for line, nextline in list_with_next_generator(self.output):
                        if line[0] >= 0 and line[2] == 3:
                            output = output + " " * int(line[0] / 2)
                        if line[1]:
                            output = output + line[1]
                        if line[2] == 2:
                            output = output + "\n"
                        elif line[2] == 4 and nextline and nextline[0] > line[0]:
                            output = output + "\n"
                        elif (
                            line[1]
                            and nextline[1]
                            and not (
                                line[1].strip().startswith("<")
                                or line[1].strip().startswith("{")
                            )
                            and not (
                                nextline[1].strip().startswith("<")
                                or nextline[1].strip().startswith("{")
                            )
                        ):
                            output = output + "\n"
                ret = output
                ret = ret.replace("|||", "\n")

            if self.output_processors and "@@(" in ret:
                tab_tmp1 = ret.split("@@(")
                ret2 = []
                for pos in tab_tmp1:
                    if ret2:
                        tab_tmp2 = pos.split(")", 1)
                        if len(tab_tmp2) > 1:
                            value = tab_tmp2[0]
                            if "://" in value:
                                fun, value = value.split("://", 1)
                            elif "-" in value:
                                fun, value = value.split("-", 1)
                            else:
                                fun = "default"
                            if fun in self.output_processors:
                                ret2.append(self.output_processors[fun](value))
                            ret2.append(tab_tmp2[1])
                        else:
                            ret2.append(pos)
                    else:
                        ret2.append(pos)

                ret = "".join(ret2)

            ret = ret.replace("\r", "").replace("\\\n", "")

            if "<inline:>" in ret:
                tab = []
                l = len(ret)
                while True:
                    ret = (
                        ret.replace(" <inline:>", "<inline:>")
                        .replace("\n<inline:>", "<inline:>")
                        .replace("</inline:> ", "</inline:>")
                        .replace("</inline:>\n", "</inline:>")
                    )
                    if len(ret) == l:
                        break
                    l = len(ret)
                x = ret.split("<inline:>")
                tab.append(x[0])
                for item in x[1:]:
                    xx = item.split("</inline:>")
                    lines = xx[0].split("\n")
                    s = " ".join([line.replace("\r", "").strip() for line in lines])
                    s = (
                        s.replace("> ", ">")
                        .replace("} ", "}")
                        .replace(" <", "<")
                        .replace(" {", "{")
                    )
                    tab.append(s)
                    tab.append(xx[1])
                ret = "".join(tab)

            return ret


def ihtml_to_html_base(file_name, input_str=None, lang="en"):
    """Convert file in ihtml format and return standard html

    Args:
        file_name - ihtml file name
        input_str - if not None - convert input_str instead input file
    Returns:
        converted html string
    """
    conwert = ConwertToHtml(file_name, ["br", "meta", "input"], [], [], input_str, lang)
    try:
        conwert.process()
        return conwert.to_str()
    except:
        import sys, traceback

        print(sys.exc_info())
        traceback.print_exc()
        return ""


def py_to_js(script, module_path):
    """Compile python script to javascript. Additional indentation are removed from python scrip before compile.

    Args:
        script - python script source
        module_path - path for target script
    """
    tab = -1
    out_tab = []
    for line in script.split("\n"):
        if tab < 0:
            if line.strip() != "":
                tab = len(line) - len(line.lstrip())
            else:
                continue
        out_tab.append(line[tab:])
    script2 = "\n".join(out_tab)
    # code = py2js(script2, module_path)
    spec_format = False
    if '"""' in script2:
        spec_format = True
        x = script2.split('"""')
        tab_script = []
        tab_string = []
        in_tabstring = False
        for pos in x:
            if in_tabstring:
                pos2 = ihtml_to_html_base(None, pos)
                tab_string.append(
                    pos2.replace("\r", "").replace("'", "\\'").replace('"', '\\"')
                )
                tab_script.append("'$$$compiled_str$$$'")
                in_tabstring = False
            else:
                tab_script.append(pos)
                in_tabstring = True
        script2 = "".join(tab_script)
        tmp = []
        for pos in tab_string:
            tmp2 = pos.split("\n")
            if len(tmp2) == 1:
                tmp.append(tmp2[0])
            elif len(tmp2) == 2:
                tmp.append(tmp2[0] + "\\n' +\n    '" + tmp2[1])
            else:
                result = tmp2[0] + "\\n' +\n"
                for pos2 in tmp2[1:-1]:
                    result += "    '" + pos2 + "\\n' +\n"
                result += "    '" + tmp2[-1]
                tmp.append(result)
        tab_string = tmp

    error, code = compile(script2)

    if error:
        print(code)
        return code
    else:
        if spec_format:
            code = code.replace('"$$$compiled_str$$$"', "'$$$compiled_str$$$'")
            x = code.split("$$$compiled_str$$$")
            result = [None] * (len(x) + len(tab_string))
            result[::2] = x
            result[1::2] = tab_string
            code = "".join(result)
        return code



from pytigon_lib.schhtml.basehtmltags import (
    BaseHtmlElemParser,
    BaseHtmlAtomParser,
    register_tag_map,
)
from pytigon_lib.schhtml.atom import Atom
from pytigon_lib.schhtml.render_helpers import (
    RenderBackground,
    RenderBorder,
    RenderMargin,
    get_size,
    sizes_from_attr,
)
from collections import deque
from .p_tags import Par
from .block_tags import BodyTag


class TdRef(object):
    def __init__(self, tdref, parent, col, col_to_parent, row_to_parent, rowspan=0):
        self.parent = parent
        self.col = col
        self.col_to_parent = col_to_parent
        self.row_to_parent = row_to_parent
        self.rowspan = rowspan
        self.colspan = 0
        self.from_ref = None
        if tdref.__class__ == TdRef:
            self.tdref = tdref.tdref
            self.from_ref = True
        else:
            self.tdref = tdref
        self.width = -1
        self.height = -1
        self.data = self.tdref.data

    def calc_height(self):
        return self.tdref.calc_height()


class TdEmptyTag(object):
    def __init__(self):
        self.colspan = 1
        self.rowspan = 1
        self.width = 0
        self.attrs = {}

    def calc_width(self, force=None):
        return [0, 0, 0]

    def calc_height(self):
        return 0

    def render(self, dc):
        dc.annotate("render_tag", {"element": self})
        if dc.handle_html_directly:
            return (0, False)
        return (0, False)

    def set_width(self, width):
        self.width = 0

    def get_width(self):
        return [self.width, self.width, self.width]

    def get_height(self):
        return 0

    def to_txt(self):
        return ""

    def to_obj_tab(self):
        return {}


class TrRef(object):
    def __init__(self, row, attrs):
        self.row = row
        self.height = -1
        self.attrs = attrs

    def __getitem__(self, id):
        return self.row[id]

    def __setitem__(self, id, value):
        self.row[id] = value

    def __len__(self):
        return len(self.row)


class TableTag(BaseHtmlAtomParser):
    def __init__(self, parent, parser, tag, attrs):
        BaseHtmlAtomParser.__init__(self, parent, parser, tag, attrs)
        self.child_tags = ["tr", "caption", "li", "ctr*"]
        self.caption = None
        if self.parent.__class__ in (BodyTag,):
            self.subtab = False
        else:
            self.subtab = True
        rb = RenderBorder(self)
        self.border = rb.get_size()
        # if "cellspacing" in attrs:
        #    self.cellspacing = sizes_from_attr(attrs["cellspacing"], self)
        # else:
        #    self.cellspacing = [0, 0, 0, 0]
        # if "cellpadding" in attrs:
        #    self.cellpadding = sizes_from_attr(attrs["cellpadding"], self)
        # else:
        #    self.cellpadding = [0, 0, 0, 0]
        # if "padding" in attrs:
        #    self.extra_space = sizes_from_attr(attrs["padding"], self)
        # else:
        #    self.extra_space = [0, 0, 0, 0]
        self.render_helpers = [RenderMargin(self), RenderBackground(self)]
        self.extra_space = get_size(self.render_helpers)

        self.tr_list = []
        self.tr_queue = deque()
        self.col_count = -1
        self.sizes = None
        self.sizes_ok = False
        self.dy_rendered = 0
        self.in_close = False
        self.start = True
        self.end = False
        self.atom = None
        self.lp = 0

    def _get_pseudo_margins(self):
        return [
            self.border[0] + self.extra_space[0],
            self.border[1] + self.extra_space[1],
            self.border[2] + self.extra_space[2],
            self.border[3] + self.extra_space[3],
        ]

    def get_width(self):
        if self.width >= 0:
            return [self.width, self.width, self.width]
        else:
            if "width" in self.attrs:
                (parent_width, parent_min, parent_max) = self.parent.get_client_width()
                width = self._norm_sizes([self.attrs["width"]], parent_width)[0]
                min = self._norm_sizes([self.attrs["width"]], parent_min)[0]
                max = self._norm_sizes([self.attrs["width"]], parent_max)[0]
                return [width, min, max]
            return self.parent.get_client_width()

    def get_client_width(self):
        (opt, min, max) = self.get_width()
        w = self.border[0] + self.border[1] + self.extra_space[0] + self.extra_space[1]
        if opt > 0:
            opt -= w
        if min > 0:
            min -= w
        if max > 0:
            max -= w
        return [opt, min, max]

    def calc_col_sizes(self):
        if self.col_count <= 0:
            return
        if not self.sizes_ok:
            if not self.sizes:
                self.sizes = [[-1, -1, -1]] * self.col_count
            for i in range(0, self.col_count):
                if self.sizes[i][0] < 0:
                    (opt, min, max) = (0, 0, 0)
                    for pos in self.tr_list:
                        if pos[i].__class__ != TdRef and pos[i].colspan == 1:
                            s = pos[i].get_width()
                            if opt < s[0]:
                                opt = s[0]
                            if min < s[1]:
                                min = s[1]
                            if max < s[2]:
                                max = s[2]
                            self.sizes[i] = [opt, min, max]
            for i in range(0, len(self.tr_list)):
                for j in range(0, self.col_count):
                    pos = self.tr_list[i][j]
                    if pos.__class__ != TdRef and pos.colspan > 1:
                        s = pos.get_width()
                        s2 = [0, 0, 0]
                        for k in range(0, pos.colspan):
                            s2[0] += self.sizes[j + k][0]
                            s2[1] += self.sizes[j + k][1]
                            s2[2] += self.sizes[j + k][2]
                        delta = [0, 0, 0]
                        for k in range(0, 3):
                            if s[k] > s2[k]:
                                delta[k] = (s[k] - s2[k]) / pos.colspan
                        if delta[0] != 0 or delta[1] != 0 or delta[2] != 0:
                            l = 1
                            if delta[l] != 0:
                                for k in range(0, pos.colspan):
                                    self.sizes[j + k][l] += delta[l]
            (opt, min, max) = (0, 0, 0)
            for pos in self.sizes:
                opt += pos[0]
                min += pos[1]
                max += pos[2]
            if self.width < 0:
                if self.parent.width >= 0:
                    parent_width = self.get_parent_width()
                    if opt > parent_width or max > parent_width:
                        self.width = parent_width
                    else:
                        self.width = (
                            opt
                            + self.border[0]
                            + self.border[1]
                            + self.extra_space[0]
                            + self.extra_space[1]
                        )
                else:
                    self.width = (
                        opt
                        + self.border[0]
                        + self.border[1]
                        + self.extra_space[0]
                        + self.extra_space[1]
                    )
            if self.width > 0:
                width2 = (
                    ((self.width - self.border[0]) - self.border[1])
                    - self.extra_space[0]
                ) - self.extra_space[1]
                if width2 < min:
                    for pos in self.sizes:
                        pos[0] = pos[1]
                else:
                    if width2 <= opt:
                        if opt - min == 0:
                            proc = 1
                        else:
                            proc = ((width2 - min) * 1.0) / (opt - min)
                        for pos in self.sizes:
                            pos[0] = pos[1] + (pos[0] - pos[1]) * proc
                    else:
                        if width2 <= max:
                            proc = ((width2 - opt) * 1.0) / (max - opt)
                            for pos in self.sizes:
                                pos[0] = pos[0] + (pos[2] - pos[0]) * proc
                        else:
                            proc = (width2 * 1.0) / max
                            for pos in self.sizes:
                                pos[0] = pos[2] * proc
            self.sizes_ok = True

    def close(self):
        self.in_close = True
        if self.subtab:
            self.calc_col_sizes()
            self.height = self.get_height()
            self.atom = Atom(self, self.width, 0, self.height, 0, None)
            self.make_atom_list()
            self.atom_list.append_atom(self.atom)
            self.parent.append_atom_list(self.atom_list)
        else:
            self.calc_col_sizes()
            BaseHtmlAtomParser.close(self)

    def calc_width(self):
        self.calc_col_sizes()
        return (self.width, self.width, self.width)

    def calc_height(self):
        self._calculate_rows_height()
        size = self._iter()
        y = 0
        for row in self.tr_list[:size]:
            sy = self._row_height(row)
            y += sy
        if self.start:
            y += self.border[2] + self.extra_space[2]
        if self.end:
            y += self.border[3] + self.extra_space[3]
        return y

    def child_ready_to_render(self, child):
        if self.dc_info.dc.handle_html_directly:
            return super().child_ready_to_render(child)

        if self.col_count < 0:
            i = 0
            for pos in child.td_list:
                i += pos.colspan
            self.col_count = i
        if len(self.tr_queue) > 0:
            row = self.tr_queue.popleft()
        else:
            row = [None] * self.col_count
        i = 0
        rowspan = 1
        for pos in child.td_list:
            try:
                while row[i] != None:
                    i += 1
            except:
                continue
            row[i] = pos
            if pos.colspan > 1:
                for j in range(1, pos.colspan):
                    row[i + j] = TdRef(pos, self, i + j, j, 0, pos.rowspan)
            i += pos.colspan
            if pos.rowspan > rowspan:
                rowspan = pos.rowspan
        row2 = []
        for pos in row:
            if pos:
                row2.append(pos)
            else:
                row2.append(TdEmptyTag())
        self.tr_list.append(TrRef(row2, child.attrs))
        if rowspan > 1:
            delta = (rowspan - len(self.tr_queue)) - 1
            if delta > 0:
                for j in range(0, delta):
                    self.tr_queue.append([None] * self.col_count)
            tr = self.tr_list[-1]
            for i in range(0, self.col_count):
                if tr[i].rowspan > 1:
                    for j in range(0, tr[i].rowspan - 1):
                        if tr[i].__class__ == TdRef:
                            col_to_parent = tr[i].col_to_parent
                        else:
                            col_to_parent = 0
                        self.tr_queue[j][i] = TdRef(
                            tr[i], self, i, col_to_parent, j + 1, 0
                        )
        if self.width >= 0 and len(self.tr_queue) == 0:
            if not self.sizes_ok and self.col_count > 0:
                if not self.sizes:
                    self.sizes = [[-1, -1, -1]] * self.col_count
                no_size_count = 0
                no_size_id = -1
                width_tab = 0
                for i in range(0, self.col_count):
                    td = self.tr_list[-1][i]
                    if td and td.width > 0:
                        self.sizes[i] = td.get_width()
                    if self.sizes[i][0] < 0:
                        no_size_count += 1
                        no_size_id = i
                    else:
                        width_tab += self.sizes[i][0]
                if no_size_count == 1:
                    self.sizes[no_size_id][0] = (
                        (((self.width - width_tab) - self.border[0]) - self.border[1])
                        - self.extra_space[0]
                    ) - self.extra_space[1]
                    self.sizes[no_size_id][1] = self.sizes[no_size_id][0]
                    self.sizes[no_size_id][2] = self.sizes[no_size_id][0]
                    self.tr_list[-1][i].set_width(self.width - width_tab)
                    no_size_count = 0
                if no_size_count == 0:
                    self.sizes_ok = True
            if self.sizes_ok and not self.subtab:
                BaseHtmlAtomParser.child_ready_to_render(self, child)
        self.lp += 1

    def _row_height(self, row):
        if row.height >= 0:
            return row.height
        else:
            dy = 0
            for i in range(0, self.col_count):
                if row[i].__class__ != TdRef and row[i].rowspan == 1:
                    width = 0
                    for j in range(0, row[i].colspan):
                        width += self.sizes[i + j][0]
                    row[i].set_width(width)
                    sy = row[i].get_height()
                    if sy > dy:
                        dy = sy
            row.height = dy
            return dy

    def _row_rowspan_height(self, rows):
        dy = 0
        for i in range(0, self.col_count):
            if rows[0][i].__class__ != TdRef and rows[0][i].rowspan > 1:
                width = 0
                for j in range(0, rows[0][i].colspan):
                    width += self.sizes[i + j][0]
                sy = rows[0][i].get_height()
                sy2 = 0
                for j in range(0, rows[0][i].rowspan):
                    sy2 += self._row_height(rows[j])
                if sy > sy2:
                    delta = (sy - sy2) / rows[0][i].rowspan
                    for j in range(0, rows[0][i].rowspan):
                        rows[j].height = rows[j].height + delta

    def _calculate_rows_height(self):
        for row in self.tr_list:
            self._row_height(row)
        for i in range(0, len(self.tr_list)):
            self._row_rowspan_height(self.tr_list[i:])

    def _iter(self):
        if self.subtab:
            self.end = True
            return len(self.tr_list)
        else:
            rows = 0
            stack = 1
            for row_id in range(0, len(self.tr_list)):
                row = self.tr_list[row_id]
                max_row_span = 1
                for td in row:
                    if td.rowspan > max_row_span:
                        max_row_span = td.rowspan
                stack += max_row_span - 2
                rows += 1
                if stack == 0 and not (
                    "page-break-inside" in row.attrs
                    and row.attrs["page-break-inside"] == "avoid"
                ):
                    if self.in_close and len(self.tr_list) == rows:
                        self.end = True
                    return rows
            if self.in_close:
                self.end = True
            return rows

    def render(self, dc_parm):
        self.reg_id(dc_parm)
        self.reg_end()

        dc_parm.annotate("render_tag", {"element": self})

        if dc_parm.handle_html_directly:
            return (0, False)

        if "border-color" in self.attrs:
            rgb = dc_parm.rgbfromhex(self.attrs["border-color"])
        else:
            rgb = (0, 0, 0)
        if not self.sizes:
            if self.border[2] > 0:
                dc_parm.set_color(rgb[0], rgb[1], rgb[2], 255)
                dc_parm.set_line_width(self.border[0])
                dc_parm.add_line(
                    self.extra_space[0],
                    self.extra_space[2],
                    (self.width - self.extra_space[0]) - self.extra_space[1],
                    0,
                )
                dc_parm.draw()
            return (self.extra_space[2] + self.extra_space[3], False)
        self._calculate_rows_height()
        dc2 = dc_parm
        for r in self.render_helpers:
            dc2 = r.render(dc2)
        width = 0
        if self.width < 0:
            for pos in self.sizes:
                width += pos[0]
            self.width = width
        if self.border[0] > 0:
            brd = self.border[0]
            brd1 = brd / 2
            brd2 = brd - brd1
        else:
            brd = brd1 = brd2 = 0
        if self.border[0] + self.extra_space[0] > 0:
            if self.border[0] > 0:
                dc_parm.set_color(rgb[0], rgb[1], rgb[2], 255)
                dc_parm.set_line_width(self.border[0])
                if self.subtab:
                    dc_parm.add_line(
                        self.extra_space[0] + brd1,
                        self.extra_space[2] + brd1,
                        0,
                        ((dc_parm.dy - self.extra_space[2]) - self.extra_space[3])
                        - brd,
                    )
                    dc_parm.add_line(
                        (dc_parm.dx - 1 * self.extra_space[1]) - brd2,
                        self.extra_space[2] + brd1,
                        0,
                        ((dc_parm.dy - self.extra_space[2]) - self.extra_space[3])
                        - brd,
                    )
                else:
                    if not self.start:
                        if self.end:
                            dc_parm.add_line(
                                self.extra_space[0] + brd1,
                                -1 * brd2,
                                0,
                                dc_parm.dy - self.extra_space[3],
                            )
                            dc_parm.add_line(
                                (self.width - 1 * self.extra_space[1]) - brd2,
                                -1 * brd2,
                                0,
                                dc_parm.dy - self.extra_space[3],
                            )
                        else:
                            dc_parm.add_line(
                                self.extra_space[0] + brd1, -1 * brd2, 0, dc_parm.dy
                            )
                            dc_parm.add_line(
                                (self.width - 1 * self.extra_space[1]) - brd2,
                                -1 * brd2,
                                0,
                                dc_parm.dy,
                            )
                dc_parm.draw()
            dc = dc_parm.subdc(
                self.border[0] + self.extra_space[0],
                0,
                (((dc_parm.dx - self.border[0]) - self.border[1]) - self.extra_space[0])
                - self.extra_space[1],
                dc_parm.dy,
            )
        else:
            dc = dc_parm
        if self.start:
            if self.border[2] > 0:
                dc_parm.set_color(rgb[0], rgb[1], rgb[2], 255)
                dc.set_line_width(self.border[0])
                dc.add_line(
                    -1 * brd2,
                    self.extra_space[2] + brd1,
                    ((self.width - self.extra_space[0]) - self.extra_space[1]) - brd,
                    0,
                )
                dc.draw()
            self.start = False
            if not self.subtab:
                self.height = -1
                return (self.border[2] + self.extra_space[2], True)
            else:
                dc = dc.subdc(
                    0,
                    self.extra_space[2] + self.border[2],
                    dc.dx,
                    dc.dy - self.extra_space[2],
                )
        size = self._iter()
        y = 0
        for row_id in range(0, size):
            row = self.tr_list[row_id]
            x = 0
            sy = self._row_height(row)
            for col in range(0, self.col_count):
                if row[col].__class__ != TdRef:
                    if row[col].colspan > 1:
                        dx = 0
                        for i in range(0, row[col].colspan):
                            dx += self.sizes[col + i][0]
                    else:
                        dx = self.sizes[col][0]
                    if row[col].rowspan > 1:
                        dy = sy
                        for i in range(1, row[col].rowspan):
                            dy += self._row_height(self.tr_list[row_id + i])
                    else:
                        dy = sy
                    row[col].render(dc.subdc(x, y, dx, dy))
                x += self.sizes[col][0]
            y += sy
        if len(self.tr_list) > size:
            self.tr_list = self.tr_list[size:]
            cont = True
        else:
            self.tr_list = []
            cont = False
        self.dy_rendered += dc.get_size()[1]
        if self.end:
            if self.border[0] > 0:
                dc_parm.set_color(rgb[0], rgb[1], rgb[2], 255)
                dc.set_line_width(self.border[0])
                dc_parm.add_line(
                    self.extra_space[0] + brd1,
                    (dc_parm.dy - self.extra_space[3]) - brd2,
                    ((self.width - self.extra_space[0]) - self.extra_space[1]) - brd,
                    0,
                )
                dc_parm.draw()
            y += self.border[3] + self.extra_space[3]
        self.height = -1
        return (y, cont)

    def draw_atom(self, dc, style, x, y, dx, dy):
        if not self.sizes:
            return
        self.calc_col_sizes()
        xx = 0
        for col in range(0, self.col_count):
            xx += self.sizes[col][0]
        dc2 = dc.subdc(x, y, self.width, self.height)
        cont = True
        while cont:
            (dy, cont) = self.render(dc2)
            dc2.y += dy


class TrTag(BaseHtmlElemParser):
    def __init__(self, parent, parser, tag, attrs):
        BaseHtmlElemParser.__init__(self, parent, parser, tag, attrs)
        self.child_tags = ["td", "th"]
        self.td_list = []

    def close(self):
        if len(self.td_list) > 0:
            BaseHtmlElemParser.close(self)

    def child_ready_to_render(self, child):
        if self.dc_info.dc.handle_html_directly:
            return super().child_ready_to_render(child)
        self.td_list.append(child)

    def get_width(self):
        return self.parent.get_width()

    def get_client_width(self):
        return self.parent.get_client_width()

    def set_width(self, value):
        pass

    def _get_pseudo_margins(self):
        return [1, 1, 1, 1]

    width = property(get_width, set_width)


class TdTag(Par):
    def __init__(self, parent, parser, tag, attrs):
        if not "border" in attrs:
            if "border" in parent.parent.attrs:
                attrs["border"] = parent.parent.attrs["border"]
        if not "margin" in attrs and "cellspacing" in parent.parent.attrs:
            attrs["margin"] = parent.parent.attrs["cellspacing"]
        if not "padding" in attrs and "cellpadding" in parent.parent.attrs:
            attrs["padding"] = parent.parent.attrs["cellpadding"]

        Par.__init__(self, parent, parser, tag, attrs)

        self.border = [0, 0, 0, 0]
        self.padding = [0, 0, 0, 0]
        if "colspan" in attrs:
            self.colspan = int(attrs["colspan"])
        else:
            self.colspan = 1
        if "rowspan" in attrs:
            self.rowspan = int(attrs["rowspan"])
        else:
            self.rowspan = 1
        if "width" in attrs:
            parent_width = parent.get_client_width()[0]
            if parent_width > 0:
                self.width = self._norm_sizes([attrs["width"]], parent_width)[0]
        self.extra_space = get_size(self.render_helpers)

    def close(self):
        BaseHtmlAtomParser.close(self)

    def get_client_width(self):
        (opt, min, max) = self.get_width()
        if opt > 0:
            opt -= self.border[0] + self.border[1] + self.padding[0] + self.padding[1]
        if min > 0:
            min -= self.border[0] + self.border[1] + self.padding[0] + self.padding[1]
        if max > 0:
            max -= self.border[0] + self.border[1] + self.padding[0] + self.padding[1]
        return [opt, min, max]


class CaptionTag(BaseHtmlElemParser):
    def __init__(
        self,
        parent,
        parser,
        tag,
        attrs,
    ):
        BaseHtmlElemParser.__init__(self, parent, parser, tag, attrs)
        self.td_list = []

    def close(self):
        self.parent.caption = "".join(self.data)
        BaseHtmlElemParser.close(self)


register_tag_map("table", TableTag)
register_tag_map("caption", CaptionTag)
register_tag_map("tr", TrTag)
register_tag_map("td", TdTag)
register_tag_map("th", TdTag)



from django.core.files.storage import default_storage
import logging

# Configure logging
logging.basicConfig(level=logging.ERROR)
logger = logging.getLogger(__name__)


def filesystemcmd(cproxy=None, **kwargs):
    """Background tasks related to file system operations.

    Args:
        cproxy: Optional proxy object to send events.
        **kwargs: Keyword arguments containing 'param' dictionary with 'cmd', 'files', and optionally 'dest'.
    """
    try:
        if cproxy:
            cproxy.send_event("start")

        param = kwargs.get("param", {})
        cmd = param.get("cmd")
        files = param.get("files", [])
        dest = param.get("dest", "") + "/" if "dest" in param else ""

        if not cmd or not files:
            raise ValueError("Missing 'cmd' or 'files' in parameters.")

        if cmd == "DELETE":
            for f in files:
                try:
                    if default_storage.fs.isfile(f):
                        default_storage.fs.remove(f)
                    else:
                        default_storage.fs.removetree(f)
                except Exception as e:
                    logger.error(f"Error deleting {f}: {e}")

        elif cmd == "COPY":
            for f in files:
                try:
                    name = f.rsplit("/", 1)[-1]
                    if default_storage.fs.isfile(f):
                        default_storage.fs.copy(f, dest + name, overwrite=True)
                    else:
                        default_storage.fs.copydir(
                            f, dest + name, overwrite=True, ignore_errors=True
                        )
                except Exception as e:
                    logger.error(f"Error copying {f}: {e}")

        elif cmd == "MOVE":
            for f in files:
                try:
                    name = f.rsplit("/", 1)[-1]
                    if default_storage.fs.isfile(f):
                        default_storage.fs.move(f, dest + name, overwrite=True)
                    else:
                        default_storage.fs.movedir(
                            f, dest + name, overwrite=True, ignore_errors=True
                        )
                except Exception as e:
                    logger.error(f"Error moving {f}: {e}")

        else:
            raise ValueError(f"Unsupported command: {cmd}")

    except ValueError as e:
        raise ValueError(e)

    except Exception as e:
        logger.error(f"Unexpected error: {e}")
    finally:
        if cproxy:
            cproxy.send_event("stop")



import io

IMAGE = None


def spec_resize(image, width=0, height=0):
    """
    Resize the image by dividing it into 9 parts and resizing each part accordingly.

    Args:
        image (PIL.Image): The image to resize.
        width (int): The desired width of the output image.
        height (int): The desired height of the output image.

    Returns:
        PIL.Image: The resized image.
    """
    try:
        w = int(image.width / 3)
        h = int(image.height / 3)

        # Define the x and y coordinates for cropping
        xtab = ((0, w), (w, image.width - w), (image.width - w, image.width))
        ytab = ((0, h), (h, image.height - h), (image.height - h, image.height))

        tab = []
        i = 0

        # Crop and resize each part of the image
        for y in ytab:
            for x in xtab:
                image2 = image.crop((x[0], y[0], x[1], y[1]))

                if i in (1, 7):  # Top and bottom middle parts
                    new_width = max(1, width - 2 * w)
                    image2 = image2.resize((new_width, h))
                elif i in (3, 5):  # Left and right middle parts
                    new_height = max(1, height - 2 * h)
                    image2 = image2.resize((w, new_height))
                elif i == 4:  # Center part
                    new_width = max(1, width - 2 * w)
                    new_height = max(1, height - 2 * h)
                    image2 = image2.resize((new_width, new_height))

                tab.append(image2)
                i += 1

        # Define the x and y coordinates for pasting
        xtab = (0, w, width - w)
        ytab = (0, h, height - h)

        # Create a new image and paste the resized parts
        global IMAGE
        if not IMAGE:
            from PIL import Image as IMAGE
        dst = IMAGE.new("RGB", (width, height))
        i = 0
        for y in ytab:
            for x in xtab:
                dst.paste(tab[i], (x, y))
                i += 1

        return dst

    except Exception as e:
        raise ValueError(f"Error during image resizing: {e}")


def svg_to_png(svg_str, width=0, height=0, image_type="simple"):
    """
    Convert an SVG string to a PNG image with optional resizing.

    Args:
        svg_str (bytes): The SVG string to convert.
        width (int): The desired width of the output image.
        height (int): The desired height of the output image.
        image_type (str): The type of image to generate ("simple", "simple_min", or "frame").

    Returns:
        bytes: The PNG image as bytes.
    """
    try:
        x = svg2rlg
    except:
        from svglib.svglib import svg2rlg
    try:
        svg_io = io.BytesIO(svg_str)
        drawing = svg2rlg(svg_io)

        if image_type in ("simple", "simple_min"):
            scale_x = scale_y = 1

            if width > 0:
                scale_x = width / drawing.width
            if height > 0:
                scale_y = height / drawing.height

            if image_type == "simple_min":
                scale_x = scale_y = min(scale_x, scale_y)
            else:
                if not scale_y and scale_x:
                    scale_y = scale_x
                elif not scale_x and scale_y:
                    scale_x = scale_y

            drawing.width *= scale_x
            drawing.height *= scale_y
            drawing.scale(scale_x, scale_y)

            return drawing.asString("png")

        else:  # image_type == "frame"
            if width or height:
                if not height:
                    height = int(drawing.height * width / drawing.width)
                if not width:
                    width = int(drawing.width * height / drawing.height)

                global IMAGE
                if not IMAGE:
                    from PIL import Image as IMAGE

                img = IMAGE.open(io.BytesIO(drawing.asString("png")))
                img2 = spec_resize(img, width, height)

                output = io.BytesIO()
                img2.save(output, "PNG")
                return output.getvalue()

            else:
                return drawing.asString("png")

    except Exception as e:
        raise ValueError(f"Error during SVG to PNG conversion: {e}")


def mse(image_array1, image_array2):
    """
    The mean squared error between two images.

    Parameters
    ----------
    image_array1, image_array2 : ndarray
        Input images.

    Returns
    -------
    err : float
        The mean squared error (MSE) between the two input images.
    """
    err = np.sum((image_array1.astype("float") - image_array2.astype("float")) ** 2)
    err /= float(image_array1.shape[0] * image_array1.shape[1])
    return err


def compare_images(img1, img2):
    """
    Compare the similarity of two images.

    Parameters
    ----------
    img1, img2 : PIL.Image
        Input images. img1 is the reference image.

    Returns
    -------
    err : float
        The mean squared error (MSE) between the two input images.
    """
    try:
        x = np
    except:
        import numpy as np

    global IMAGE
    if not IMAGE:
        from PIL import Image as IMAGE

    np.array(img1), np.array(img2)
    img2_mod = img2.convert("RGB").resize(
        (img1.size[0], img1.size[1]), IMAGE.Resampling.LANCZOS
    )
    return mse(np.array(img1.convert("RGB")), np.array(img2_mod))



import io
from math import pi
import cairo

from pytigon_lib.schhtml.basedc import BaseDc, BaseDcInfo


class CairoDc(BaseDc):
    def __init__(
        self,
        ctx=None,
        calc_only=False,
        width=-1,
        height=-1,
        output_name=None,
        output_stream=None,
        scale=1.0,
        notify_callback=None,
        record=False,
    ):
        super().__init__(
            calc_only,
            width,
            height,
            output_name,
            output_stream,
            scale,
            notify_callback,
            record,
        )
        self.width = width if width >= 0 else self.default_width
        self.height = height if height >= 0 else self.default_height
        self.dc_info = CairoDcInfo(self)
        self.type = None
        self.last_style_tab = None
        self.last_move_to = None

        if self.calc_only:
            self.surf = cairo.ImageSurface(cairo.FORMAT_RGB24, 10, 10)
            self.ctx = cairo.Context(self.surf)
        else:
            if ctx:
                self.surf = None
                self.ctx = ctx
            else:
                self._initialize_surface(output_name, output_stream)
            self.ctx.set_line_cap(cairo.LINE_CAP_ROUND)

    def _initialize_surface(self, output_name, output_stream):
        if output_name:
            name = output_name.lower()
            if ".pdf" in name:
                self.surf = cairo.PDFSurface(
                    output_stream or output_name, self.width, self.height
                )
                self.type = "pdf"
            elif ".svg" in name:
                self.surf = cairo.SVGSurface(
                    output_stream or output_name, self.width, self.height
                )
                self.type = "svg"
            elif ".png" in name:
                self.surf = cairo.ImageSurface(
                    cairo.FORMAT_RGB24, self.width, self.height
                )
                self.type = "png"
            else:
                self.surf = cairo.ImageSurface(
                    cairo.FORMAT_RGB24, self.width, self.height
                )
        else:
            self.surf = cairo.ImageSurface(cairo.FORMAT_RGB24, self.width, self.height)
        self.ctx = cairo.Context(self.surf)

    def close(self):
        if not self.calc_only and self.surf:
            if self.type in ("svg", "pdf"):
                self.ctx.show_page()
            self.surf.finish()

    def _move_to(self, x, y):
        self.ctx.move_to(x * self.scale, y * self.scale)
        self.last_move_to = (x * self.scale, y * self.scale)

    def start_page(self):
        self.ctx.show_page()
        super().start_page()

    def draw(self, preserve=False):
        self.ctx.stroke_preserve() if preserve else self.ctx.stroke()
        super().draw(preserve)

    def fill(self, preserve=False):
        self.ctx.fill_preserve() if preserve else self.ctx.fill()
        super().fill()

    def set_color(self, r, g, b, a=255):
        self.ctx.set_source_rgb(r / 256.0, g / 256.0, b / 256.0)
        super().set_color(r, g, b, a)

    def set_line_width(self, width):
        self.ctx.set_line_width(width * self.scale)
        super().set_line_width(width)

    def set_style(self, style):
        if style == self.last_style:
            return self.last_style_tab
        style_tab = self.dc_info.styles[style].split(";")
        self.last_style_tab = style_tab
        slant = (
            cairo.FONT_SLANT_ITALIC if style_tab[3] == "1" else cairo.FONT_SLANT_NORMAL
        )
        weight = (
            cairo.FONT_WEIGHT_BOLD if style_tab[4] == "1" else cairo.FONT_WEIGHT_NORMAL
        )
        self.ctx.select_font_face(style_tab[1], slant, weight)
        r, g, b = self.rgbfromhex(style_tab[0])
        self.ctx.set_font_size(
            (self.scale * self.base_font_size * int(style_tab[2])) / 100
        )
        self.ctx.set_source_rgb(r / 256.0, g / 256.0, b / 256.0)
        super().set_style(style)
        return style_tab

    def add_line(self, x, y, dx, dy):
        self._move_to(x, y)
        self.ctx.line_to((x + dx) * self.scale, (y + dy) * self.scale)
        self.last_move_to = ((x + dx) * self.scale, (y + dy) * self.scale)
        super().add_line(x, y, dx, dy)

    def add_rectangle(self, x, y, dx, dy):
        self.ctx.rectangle(
            x * self.scale, y * self.scale, dx * self.scale, dy * self.scale
        )
        super().add_rectangle(x, y, dx, dy)

    def add_rounded_rectangle(self, x, y, dx, dy, radius):
        degrees = pi / 180.0
        self.ctx.new_sub_path()
        self.ctx.arc(
            (x + dx - radius) * self.scale,
            (y + radius) * self.scale,
            radius * self.scale,
            -90 * degrees,
            0 * degrees,
        )
        self.ctx.arc(
            (x + dx - radius) * self.scale,
            (y + dy - radius) * self.scale,
            radius * self.scale,
            0 * degrees,
            90 * degrees,
        )
        self.ctx.arc(
            (x + radius) * self.scale,
            (y + dy - radius) * self.scale,
            radius * self.scale,
            90 * degrees,
            180 * degrees,
        )
        self.ctx.arc(
            (x + radius) * self.scale,
            (y + radius) * self.scale,
            radius * self.scale,
            180 * degrees,
            270 * degrees,
        )
        self.ctx.close_path()
        super().add_rounded_rectangle(x, y, dx, dy, radius)

    def add_arc(self, x, y, radius, angle1, angle2):
        self.ctx.arc(
            x * self.scale,
            y * self.scale,
            radius * self.scale,
            (2 * pi * angle1) / 360,
            (2 * pi * angle2) / 360,
        )
        super().add_arc(x, y, radius, angle1, angle2)

    def add_ellipse(self, x, y, dx, dy):
        self.ctx.save()
        self.ctx.translate((x + dx / 2) * self.scale, (y + dy / 2) * self.scale)
        self.ctx.scale(self.scale * dx / 2.0, self.scale * dy / 2.0)
        self.ctx.arc(0.0, 0.0, 1.0, 0.0, 2.0 * pi)
        self.ctx.restore()
        super().add_ellipse(x, y, dx, dy)

    def add_polygon(self, xytab):
        pos0 = xytab[0]
        self.ctx.move_to(pos0[0] * self.scale, pos0[1] * self.scale)
        for pos in xytab[1:]:
            self.ctx.line_to(pos[0] * self.scale, pos[1] * self.scale)
        self.ctx.close_path()
        self.last_move_to = [pos[0] * self.scale, pos[1] * self.scale]
        super().add_polygon(xytab)

    def add_spline(self, xytab, close):
        pos0 = xytab[0]
        self.ctx.move_to(pos0[0], pos0[1])
        for pos in xytab[1:]:
            self.ctx.line_to(pos[0], pos[1])
            self.last_move_to = pos
        if close:
            self.ctx.close_path()
            self.last_move_to = pos0
        super().add_spline(xytab, close)

    def draw_text(self, x, y, txt):
        sizes = self.ctx.text_extents(txt)[:4]
        self.ctx.move_to((x - sizes[0] / 2) * self.scale, y * self.scale)
        self.ctx.show_text(txt)
        super().draw_text(x, y, txt)

    def draw_rotated_text(self, x, y, txt, angle):
        self.ctx.save()
        self.ctx.move_to(x * self.scale, y * self.scale)
        self.ctx.rotate((2 * pi * angle) / 360)
        self.ctx.show_text(txt)
        self.ctx.restore()
        super().draw_rotated_text(x, y, txt, angle)

    def draw_image(self, x, y, dx, dy, scale, png_data):
        try:
            png_stream = io.BytesIO(png_data)
            surface = cairo.ImageSurface.create_from_png(png_stream)
        except Exception:
            surface = cairo.ImageSurface.create_from_png("sleeptimer.png")
        w = surface.get_width()
        h = surface.get_height()
        self.ctx.save()
        self.ctx.rectangle(x, y, dx, dy)
        self.ctx.clip()
        x_scale, y_scale = self._scale_image(x, y, dx, dy, scale, w, h)
        if scale < 4:
            self.ctx.scale(x_scale, y_scale)
            self.ctx.set_source_surface(surface, x / x_scale, y / y_scale)
            self.ctx.paint()
        else:
            delta_x = 0
            delta_y = 0
            while delta_y < dy:
                if scale == 4 and delta_y > 0:
                    break
                delta_x = 0
                while delta_x < dx:
                    if scale == 5 and delta_x > 0:
                        break
                    self.ctx.set_source_surface(surface, x + delta_x, y + delta_y)
                    self.ctx.paint()
                    delta_x += w
                delta_y += h
        self.ctx.restore()
        super().draw_image(x, y, dx, dy, scale, png_data)


class CairoDcInfo(BaseDcInfo):
    def __init__(self, dc):
        super().__init__(dc)

    def get_line_dy(self, height):
        return height * 12

    def get_extents(self, word, style):
        self.dc.set_style(style)
        sizes = self.dc.ctx.text_extents(word + ".")[:4]
        dx = sizes[2]
        dy_up = -1 * sizes[1]
        dy_down = sizes[3] - dy_up
        sizes2 = self.dc.ctx.text_extents(".")[:4]
        dx_space = sizes2[2]
        dx -= dx_space
        if word[-1] != " ":
            dx_space = 0
        return dx, dx_space, dy_up, dy_down

    def get_text_width(self, txt, style):
        self.dc.set_style(style)
        x_off, y_off, tw, th = self.dc.ctx.text_extents(txt)[:4]
        return tw

    def get_text_height(self, txt, style):
        self.dc.set_style(style)
        x_off, y_off, tw, th = self.dc.ctx.text_extents(txt)[:4]
        return th

    def get_img_size(self, png_data):
        try:
            png_stream = io.BytesIO(png_data)
            surface = cairo.ImageSurface.create_from_png(png_stream)
        except Exception:
            surface = None
        if surface:
            return surface.get_width(), surface.get_height()
        return 0, 0


def get_PdfCairoDc(result, width, height):
    surf = cairo.PDFSurface(result, width, height)
    ctx = cairo.Context(surf)
    return CairoDc(ctx=ctx, calc_only=False, width=width, height=height)



import email
from django.core.mail import EmailMultiAlternatives
from django.template.loader import get_template
from django.template import Template, Context
from email.mime.image import MIMEImage


class PytigonEmailMessage(EmailMultiAlternatives):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.html_body = None

    def set_html_body(self, context, html_template_name, txt_template_name=None):
        """Set the HTML and plain text body of the email."""
        try:
            template_html = get_template(html_template_name)
            txt_template_name = txt_template_name or html_template_name.replace(
                ".html", ".txt"
            )
            template_plain = get_template(txt_template_name)

            self.html_body = template_html.render(context)
            self.body = template_plain.render(context)
            self.attach_alternative(self.html_body, "text/html")
        except Exception as e:
            raise ValueError(f"Error setting HTML body: {e}")

    def _process_part(self, part):
        """Process each part of the email message."""
        if part.get_content_maintype() == "multipart":
            for item in part.get_payload():
                self._process_part(item)
        elif part.get_content_maintype() == "text" and not self.html_body:
            encoding = (
                part.get("Content-Type", "").split('"')[1]
                if '"' in part.get("Content-Type", "")
                else "utf-8"
            )
            if part.get_content_type() == "text/plain":
                self.body = part.get_payload(decode=True).decode(encoding)
            else:
                self.attach_alternative(
                    part.get_payload(decode=True).decode(encoding),
                    part.get_content_type(),
                )
                self.html_body = "OK"
        elif part.get_content_maintype() == "image":
            img = MIMEImage(part.get_payload(decode=True))
            for key, value in part.items():
                img.add_header(key, value)
            self.attach(img)
        else:
            self.attach(part)

    def set_eml_body(self, context, eml_template_name):
        """Set the email body from an EML template."""
        try:
            template_eml = get_template(eml_template_name)
            eml_name = template_eml.origin.name
            with open(eml_name, "rt") as f:
                t = Template(f.read())
                c = Context(context)
                txt = t.render(c)
                self._process_part(email.message_from_string(txt))
        except Exception as e:
            raise ValueError(f"Error setting EML body: {e}")


def send_message(
    subject,
    message_template_name,
    from_email,
    to,
    bcc=None,
    context=None,
    message_txt_template_name=None,
    prepare_message=None,
    send=True,
):
    """Send an email message."""
    if context is None:
        context = {}

    message = PytigonEmailMessage(subject, "", from_email, to, bcc)

    try:
        if message_template_name.endswith(".html"):
            message.set_html_body(
                context, message_template_name, message_txt_template_name
            )
        elif message_template_name.endswith(".eml"):
            message.set_eml_body(context, message_template_name)

        if prepare_message:
            prepare_message(message)

        if send:
            message.send()

        return message
    except Exception as e:
        raise ValueError(f"Error sending message: {e}")



import os
import environ
from typing import Optional

# Global environment variable instance
ENV = None


def get_environ(path: Optional[str] = None) -> environ.Env:
    """Initialize and return the environment configuration.

    Args:
        path (Optional[str]): Path to the directory containing .env or env file.

    Returns:
        environ.Env: The environment configuration instance.
    """
    global ENV

    if ENV is None:
        ENV = environ.Env(
            DEBUG=(bool, False),
            PYTIGON_DEBUG=(bool, False),
            EMBEDED_DJANGO_SERVER=(bool, False),
            PYTIGON_WITHOUT_CHANNELS=(bool, False),
            PYTIGON_TASK=(bool, False),
            LOGS_TO_DOCKER=(bool, False),
            PWA=(bool, False),
            GRAPHQL=(bool, False),
            DJANGO_Q=(bool, False),
            ALLAUTH=(bool, False),
            REST=(bool, False),
            CANCAN_ENABLED=(bool, False),
            SENTRY_ENABLED=(bool, False),
            PROMETHEUS_ENABLED=(bool, False),
            COMPRESS_ENABLED=(bool, False),
            SECRET_KEY=(str, ""),
            CHANNELS_REDIS=(str, ""),
            PUBLISH_IN_SUBFOLDER=(str, ""),
            THUMBNAIL_PROTECTED=(bool, False),
            MAILER=(bool, True),
            LOG_VIEWER=(bool, False),
            SCRIPT_MODE=(bool, False),
        )

    if path:
        env_paths = [os.path.join(path, ".env"), os.path.join(path, "env")]
        for env_path in env_paths:
            if os.path.exists(env_path):
                try:
                    ENV.read_env(env_path)
                except Exception as e:
                    print(f"Error reading environment file {env_path}: {e}")

    return ENV



from pytigon_lib.schhtml.basedc import BaseDc, BaseDcInfo
import io
from docx import Document
from docx.shared import Inches, Pt, RGBColor
import docx.enum.text

IMAGE = None


class DocxDc(BaseDc):
    def __init__(
        self,
        ctx=None,
        calc_only=False,
        width=8.5,
        height=11,
        output_name=None,
        output_stream=None,
        scale=1.0,
        notify_callback=None,
        docx_template_path=None,
        record=False,
    ):
        super().__init__(
            calc_only,
            -1,
            -1,
            output_name,
            output_stream,
            scale,
            notify_callback,
            record,
        )
        self.dc_info = DocxDcinfo(self)
        self.type = None

        self.width = width if width >= 0 else -1
        self.height = height if height >= 0 else 1000000000

        self.last_style_tab = None
        self.handle_html_directly = True
        self.document = Document(docx_template_path)

        self.page_width = width
        self.page_height = height
        self.set_margins(0.5, 0.5, 0.5, 0.5)

        self.map = {
            "body": self.body,
            "p": self.p,
            "div": self.div,
            "h1": self.h1,
            "h2": self.h2,
            "h3": self.h3,
            "h4": self.h4,
            "h5": self.h5,
            "h6": self.h6,
            "table": self.table,
            "tr": self.tr,
            "td": self.td,
            "th": self.th,
        }
        self.last_ = None

        if self.notify_callback:
            self.notify_callback("start", {"dc": self})

    def set_margins(self, top, right, bottom, left):
        current_section = self.document.sections[-1]
        current_section.top_margin = Inches(top)
        current_section.right_margin = Inches(right)
        current_section.bottom_margin = Inches(bottom)
        current_section.left_margin = Inches(left)
        self.body_width = (
            self.page_width
            - current_section.left_margin.inches
            - current_section.right_margin.inches
        )
        self.body_height = (
            self.page_height
            - current_section.top_margin.inches
            - current_section.bottom_margin.inches
        )

    def close(self):
        if self.notify_callback:
            self.notify_callback("end", {"dc": self})
        try:
            if self.output_stream:
                self.document.save(self.output_stream)
            elif self.output_name:
                self.document.save(self.output_name)
        except Exception as e:
            raise RuntimeError(f"Failed to save document: {e}")

    def annotate(self, what, data):
        if what == "end_tag":
            element = data["element"]
            parent = element.parent
            if element and parent and element.tag in self.map:
                self.map[element.tag](element, parent)

    def _handle_width_and_height(self, element):
        width = height = None
        if "width" in element.attrs:
            w = element.attrs["width"]
            width = (
                (self.body_width - 0.2) * int(w.replace("%", "")) / 100
                if "%" in w
                else int(w.replace("px", "").replace("rem", "").replace("em", "")) / 300
            )
        if "height" in element.attrs:
            h = element.attrs["height"]
            height = (
                self.body_height * int(h.replace("%", "")) / 100
                if "%" in h
                else int(h.replace("px", "").replace("rem", "").replace("em", "")) / 300
            )
        return width, height

    def _inch_from_param(self, param, orientation):
        if "%" in param:
            return (
                (self.body_height if orientation == 0 else self.body_width)
                * int(param.replace("%", ""))
                / 100
            )
        return int(param.replace("px", "").replace("rem", "").replace("em", "")) / 300

    def _add_style(self, dest_element, source_element):
        if "classes" in source_element.attrs:
            for attr in source_element.attrs["classes"].split(" "):
                if attr.startswith("Style-"):
                    dest_element.style = attr[6:].replace("-", " ")

        padding = (
            source_element.attrs.get("padding", "").split(" ")
            if "padding" in source_element.attrs
            else None
        )
        margin = (
            source_element.attrs.get("margin", "").split(" ")
            if "margin" in source_element.attrs
            else None
        )

        if hasattr(dest_element, "paragraph_format") and (padding or margin):
            dest_element.paragraph_format.left_indent = Inches(0)
            dest_element.paragraph_format.right_indent = Inches(0)
            dest_element.paragraph_format.space_before = Inches(0)
            dest_element.paragraph_format.space_after = Inches(0)

            for tab in (margin, padding):
                if tab:
                    if len(tab) == 4:
                        dest_element.paragraph_format.left_indent += Inches(
                            self._inch_from_param(tab[3], 1)
                        )
                        dest_element.paragraph_format.right_indent += Inches(
                            self._inch_from_param(tab[1], 1)
                        )
                        dest_element.paragraph_format.space_before += Inches(
                            self._inch_from_param(tab[0], 0)
                        )
                        dest_element.paragraph_format.space_after += Inches(
                            self._inch_from_param(tab[2], 0)
                        )
                    elif len(tab) == 2:
                        dest_element.paragraph_format.left_indent += Inches(
                            self._inch_from_param(tab[1], 1)
                        )
                        dest_element.paragraph_format.right_indent += Inches(
                            self._inch_from_param(tab[1], 1)
                        )
                        dest_element.paragraph_format.space_before += Inches(
                            self._inch_from_param(tab[0], 0)
                        )
                        dest_element.paragraph_format.space_after += Inches(
                            self._inch_from_param(tab[0], 0)
                        )
                    else:
                        dest_element.paragraph_format.left_indent += Inches(
                            self._inch_from_param(tab[0], 1)
                        )
                        dest_element.paragraph_format.right_indent += Inches(
                            self._inch_from_param(tab[0], 1)
                        )
                        dest_element.paragraph_format.space_before += Inches(
                            self._inch_from_param(tab[0], 0)
                        )
                        dest_element.paragraph_format.space_after += Inches(
                            self._inch_from_param(tab[0], 0)
                        )

        align = source_element.attrs.get(
            "align", source_element.attrs.get("text-align", "")
        )
        if align == "center":
            dest_element.alignment = docx.enum.text.WD_ALIGN_PARAGRAPH.CENTER
        elif align == "right":
            dest_element.alignment = docx.enum.text.WD_ALIGN_PARAGRAPH.RIGHT
        else:
            dest_element.alignment = docx.enum.text.WD_ALIGN_PARAGRAPH.LEFT

    def _add_image(self, img, dest_element, width=None, height=None):
        img_stream = io.BytesIO(img)
        dest_element.add_picture(
            img_stream,
            width=Inches(width) if width else None,
            height=Inches(height) if height else None,
        )

    def _process_atom_list(self, dest_element, source_element):
        if source_element.atom_list and source_element.atom_list.atom_list:
            for atom in source_element.atom_list.atom_list:
                if atom.style:
                    style = self.dc_info.styles[atom.style]
                    (
                        color,
                        font_family,
                        font_size,
                        font_style,
                        font_weight,
                        text_decoration,
                    ) = style.split(";")
                else:
                    style = None
                if type(atom).__name__ == "BrAtom":
                    dest_element.add_run("\n")
                elif isinstance(atom.data, str):
                    s = atom.data.replace("»", " ")
                    x = dest_element.add_run(s)
                    if style:
                        x.font.bold = int(font_weight) > 0
                        x.font.italic = int(font_style) == 1
                        x.font.size = Pt(int(int(font_size) * 10 / 100))
                        x.font.color.rgb = RGBColor(*self.rgbfromhex(color))
                elif type(atom.data).__name__ == "ImgDraw":
                    width, height = self._handle_width_and_height(atom.data.img_tag)
                    self._add_image(
                        atom.data.image, dest_element.add_run(), width, height
                    )
                else:
                    self._process_atom_list(dest_element, atom.data)

    def h(self, element, level):
        hh = self.document.add_heading("", level)
        self._add_style(hh, element)
        self._process_atom_list(hh, element)

    def p(self, element, parent):
        par = self.document.add_paragraph("")
        self._add_style(par, element)
        self._process_atom_list(par, element)

    def div(self, element, parent):
        par = self.document.add_paragraph("")
        par.paragraph_format.left_indent = 0
        par.paragraph_format.right_indent = 0
        self._add_style(par, element)
        self._process_atom_list(par, element)

    def body(self, element, parent):
        pass

    def h1(self, element, parent):
        return self.h(element, 0)

    def h2(self, element, parent):
        return self.h(element, 1)

    def h3(self, element, parent):
        return self.h(element, 2)

    def h4(self, element, parent):
        return self.h(element, 3)

    def h5(self, element, parent):
        return self.h(element, 4)

    def h6(self, element, parent):
        return self.h(element, 5)

    def table(self, element, parent):
        if element.tr_list:
            table = self.document.add_table(
                rows=len(element.tr_list), cols=len(element.tr_list[0].td_list)
            )
            self._add_style(table, element)
            for i, row in enumerate(element.tr_list):
                row_dest = table.rows[i].cells
                for j, td in enumerate(row.td_list):
                    try:
                        c = row_dest[j]
                        c._tc.clear_content()
                        p = c.add_paragraph(None)
                        self._add_style(p, td)
                        self._process_atom_list(p, td)
                    except Exception:
                        pass
            element.tr_list = []

    def tr(self, element, parent):
        if parent.tag == "table":
            parent.tr_list.append(element)
        else:
            parent.parent.tr_list.append(element)

    def td(self, element, parent):
        parent.td_list.append(element)

    def th(self, element, parent):
        parent.td_list.append(element)


class DocxDcinfo(BaseDcInfo):
    def __init__(self, dc):
        super().__init__(dc)

    def get_text_height(self, word, style):
        return 1

    def get_img_size(self, png_data):
        global IMAGE
        if not IMAGE:
            from PIL import Image as IMAGE
        try:
            png_stream = io.BytesIO(png_data)
            image = IMAGE.open(png_stream)
            return image.size
        except Exception:
            return (0, 0)



from pytigon_lib.schhtml.basehtmltags import BaseHtmlElemParser, register_tag_map
from .tags.block_tags import *
from .tags.table_tags import *
from .tags.p_tags import *
from .tags.css_tags import *
from .tags.atom_tags import *
from .tags.page_tags import *
from .tags.extra_tags import *


class HtmlTag(BaseHtmlElemParser):
    """Represents the HTML tag in the document."""

    def __init__(self, parent, parser, tag, attrs):
        """Initialize the HTML tag with default attributes."""
        super().__init__(parent, parser, tag, attrs)
        self.child_tags = ["head", "body", "script"]
        self.width = 2480
        self.height = 3508
        self.y = 0
        self.dc = None
        self.attrs = {
            "color": "#000",
            "font-family": "sans-serif",
            "font-size": "100%",
            "font-style": "normal",
            "font-weight": "normal",
            "text-decoration": "none",
        }

    def handle_data(self, data):
        """Handle data within the HTML tag."""
        pass

    def close(self):
        """Close the HTML tag."""
        super().close()

    def set_dc(self, dc):
        """Set the drawing context and update width and height."""
        try:
            self.width, self.height = dc.get_size()
            self.dc = dc
        except AttributeError as e:
            raise ValueError("Invalid drawing context provided.") from e


class HeaderTag(BaseHtmlElemParser):
    """Represents the HEAD tag in the document."""

    def __init__(self, parent, parser, tag, attrs):
        """Initialize the HEAD tag with child tags."""
        super().__init__(parent, parser, tag, attrs)
        self.child_tags = ["style", "link"]

    def close(self):
        """Close the HEAD tag."""
        pass


class CommentTag(BaseHtmlElemParser):
    """Represents a comment in the document."""

    def __init__(self, parent, parser, tag, attrs):
        """Initialize the comment tag."""
        super().__init__(parent, parser, tag, attrs)

    def handle_starttag(self, parser, tag, attrs):
        """Handle the start tag within the comment."""
        return None

    def close(self):
        """Close the comment tag."""
        pass


# Register the tags with the parser
register_tag_map("html", HtmlTag)
register_tag_map("head", HeaderTag)
register_tag_map("comment", CommentTag)



from django.http import HttpResponse, JsonResponse
from django.forms.models import model_to_dict

# HTML templates for responses
_NEW_ROW_OK_HTML = (
    """<head><meta name="RETURN" content="$$RETURN_NEW_ROW_OK" /></head>"""
)
_UPDATE_ROW_OK_HTML = (
    """<head><meta name="RETURN" content="$$RETURN_UPDATE_ROW_OK" /></head>"""
)
_DELETE_ROW_OK_HTML = """<head><meta name="RETURN" content="$$RETURN_OK" /></head>"""
_OK_HTML = """<head><meta name="RETURN" content="$$RETURN_OK" /></head>"""
_REFRESH_HTML = """<head><meta name="RETURN" content="$$RETURN_REFRESH" /></head>"""
_REFRESH_PARENT_HTML = (
    """<head><meta name="RETURN" content="$$RETURN_REFRESH_PARENT" /></head>"""
)
_RELOAD_HTML = (
    """<head><meta name="RETURN" content="$$RETURN_RELOAD" /></head><body>%s</body>"""
)
_CANCEL_HTML = """<head><meta name="RETURN" content="$$RETURN_CANCEL" /></head>"""
_ERROR_HTML = (
    """<head><meta name="RETURN" content="$$RETURN_ERROR" /></head><body>%s</body>"""
)


def _is_python_agent(request):
    """Check if the request is from a Python-based user agent."""
    user_agent = request.META.get("HTTP_USER_AGENT", "")
    return user_agent.lower().startswith("py")


def new_row_ok(request, id, obj):
    """Handle response for a successful new row creation."""
    try:
        if _is_python_agent(request):
            return JsonResponse({"action": "new_row_ok", "obj": model_to_dict(obj)})
        return HttpResponse(_NEW_ROW_OK_HTML + "id:" + str(id))
    except Exception as e:
        return HttpResponse(_ERROR_HTML % str(e), status=500)


def update_row_ok(request, id, obj):
    """Handle response for a successful row update."""
    try:
        if _is_python_agent(request):
            return JsonResponse({"action": "update_row_ok", "obj": model_to_dict(obj)})
        return HttpResponse(_UPDATE_ROW_OK_HTML + "id:" + str(id))
    except Exception as e:
        return HttpResponse(_ERROR_HTML % str(e), status=500)


def delete_row_ok(request, id, obj):
    """Handle response for a successful row deletion."""
    try:
        if _is_python_agent(request):
            return JsonResponse({"action": "delete_row_ok", "obj": model_to_dict(obj)})
        return HttpResponse(_DELETE_ROW_OK_HTML + "id:" + str(id))
    except Exception as e:
        return HttpResponse(_ERROR_HTML % str(e), status=500)


def ok(request):
    """Handle a generic OK response."""
    return HttpResponse(_OK_HTML)


def refresh(request):
    """Handle a refresh response."""
    return HttpResponse(_REFRESH_HTML)


def refresh_parent(request):
    """Handle a refresh parent response."""
    return HttpResponse(_REFRESH_PARENT_HTML)


def reload(request, new_html):
    """Handle a reload response with new HTML content."""
    return HttpResponse(_RELOAD_HTML % new_html)


def cancel(request):
    """Handle a cancel response."""
    return HttpResponse(_CANCEL_HTML)


def error(request, error_txt):
    """Handle an error response with the provided error text."""
    return HttpResponse(_ERROR_HTML % error_txt, status=400)



from django import forms
from pytigon_lib.schdjangoext.formwidgets import (
    CheckboxSelectMultipleWithIcon,
    RadioSelectWithIcon,
)
from django_select2.forms import (
    Select2Widget,
    Select2MultipleWidget,
    HeavySelect2Widget,
    HeavySelect2MultipleWidget,
    ModelSelect2Widget,
    ModelSelect2MultipleWidget,
)


class ModelChoiceFieldWithIcon(forms.ModelChoiceField):
    """Extended version of django's ModelChoiceField.
    If the label contains '|', it splits into two parts: the first part is the image address,
    and the second part is the label.
    """

    widget = RadioSelectWithIcon


class ModelMultipleChoiceFieldWithIcon(forms.ModelMultipleChoiceField):
    """Extended version of django's ModelMultipleChoiceField with icon support."""

    widget = CheckboxSelectMultipleWithIcon


class Select2Field(forms.ChoiceField):
    """A ChoiceField that uses Select2Widget for better UI."""

    def __init__(self, choices=(), attrs=None, **kwargs):
        attrs = attrs or {}
        attrs.setdefault("data-minimum-input-length", 0)
        widget = Select2Widget(attrs=attrs, choices=choices)
        super().__init__(choices=choices, widget=widget, **kwargs)


class Select2MultipleField(forms.MultipleChoiceField):
    """A MultipleChoiceField that uses Select2MultipleWidget for better UI."""

    def __init__(self, choices=(), attrs=None, **kwargs):
        attrs = attrs or {}
        attrs.setdefault("data-minimum-input-length", 0)
        widget = Select2MultipleWidget(attrs=attrs, choices=choices)
        super().__init__(choices=choices, widget=widget, **kwargs)


class HeavySelect2Field(forms.ChoiceField):
    """A ChoiceField that uses HeavySelect2Widget for AJAX-based options."""

    def __init__(self, data_url, attrs=None, **kwargs):
        attrs = attrs or {}
        attrs.setdefault("data-minimum-input-length", 0)
        widget = HeavySelect2Widget(data_url=data_url, attrs=attrs)
        super().__init__(widget=widget, **kwargs)


class HeavySelect2MultipleField(forms.MultipleChoiceField):
    """A MultipleChoiceField that uses HeavySelect2MultipleWidget for AJAX-based options."""

    def __init__(self, data_url, attrs=None, **kwargs):
        attrs = attrs or {}
        attrs.setdefault("data-minimum-input-length", 0)
        widget = HeavySelect2MultipleWidget(data_url=data_url, attrs=attrs)
        super().__init__(widget=widget, **kwargs)


class ModelSelect2Field(forms.ModelChoiceField):
    """A ModelChoiceField that uses ModelSelect2Widget for AJAX-based model choices."""

    def __init__(
        self,
        model=None,
        queryset=None,
        search_fields=None,
        attrs=None,
        empty_label="-----",
        **kwargs
    ):
        attrs = attrs or {}
        attrs.setdefault("data-minimum-input-length", 0)
        widget = ModelSelect2Widget(
            model=model,
            queryset=queryset,
            search_fields=search_fields,
            empty_label=empty_label,
            attrs=attrs,
        )
        widget.attrs["style"] = "width:100%;"
        super().__init__(widget=widget, queryset=queryset, **kwargs)


class ModelSelect2MultipleField(forms.ModelMultipleChoiceField):
    """A ModelMultipleChoiceField that uses ModelSelect2MultipleWidget for AJAX-based model choices."""

    def __init__(
        self,
        model=None,
        queryset=None,
        search_fields=None,
        attrs=None,
        empty_label="-----",
        **kwargs
    ):
        attrs = attrs or {}
        attrs.setdefault("data-minimum-input-length", 0)
        widget = ModelSelect2MultipleWidget(
            model=model,
            queryset=queryset,
            search_fields=search_fields,
            empty_label=empty_label,
            attrs=attrs,
        )
        super().__init__(widget=widget, queryset=queryset, **kwargs)



import platform
from os import environ


def platform_name():
    """Determine the platform name.

    Returns:
        str: The name of the platform (e.g., 'Linux', 'Android', 'Windows', etc.).
    """
    try:
        system_name = platform.system()
        if system_name == "Linux" and "ANDROID_ARGUMENT" in environ:
            return "Android"
        return system_name
    except Exception as e:
        print(f"Error determining platform: {e}")
        return "Unknown"



from pytigon_lib.schhtml.basehtmltags import BaseHtmlAtomParser, register_tag_map
from pytigon_lib.schhtml.atom import Atom
from pytigon_lib.schhtml.render_helpers import (
    RenderBackground,
    RenderBorder,
    RenderCellSpacing,
    RenderCellPadding,
    get_size,
)


class VectorImg(BaseHtmlAtomParser):
    """A class to handle vector image tags in HTML."""

    def __init__(self, parent, parser, tag, attrs):
        """Initialize the VectorImg object.

        Args:
            parent: The parent HTML element.
            parser: The HTML parser.
            tag: The HTML tag.
            attrs: The attributes of the HTML tag.
        """
        super().__init__(parent, parser, tag, attrs)
        self.gparent = parent.gparent
        self.render_helpers = [
            RenderCellSpacing(self),
            RenderBorder(self),
            RenderBackground(self),
            RenderCellPadding(self),
        ]
        self.extra_space = get_size(self.render_helpers)
        self.draw_txt = ""
        self.dx = 0
        self.dy = 0

    def _get_pseudo_margins(self):
        """Get the pseudo margins for the element.

        Returns:
            list: A list of margin values [top, right, bottom, left].
        """
        return list(self.extra_space)

    def close(self):
        """Close the VectorImg element and finalize its properties."""
        try:
            if self.width > 0 and self.height > 0:
                self.dx = self.width
                self.dy = self.height
            else:
                self.dx = 100
                self.dy = 100

            img_atom = Atom(self, self.dx, 0, self.dy, 0)
            img_atom.set_parent(self)
            self.make_atom_list()
            self.atom_list.append_atom(img_atom)
            self.parent.append_atom_list(self.atom_list)
        except Exception as e:
            raise RuntimeError(f"Error closing VectorImg: {e}")

    def handle_data(self, data):
        """Handle the data within the VectorImg element.

        Args:
            data: The data to be handled.
        """
        self.draw_txt += data

    def draw_atom(self, dc, style, x, y, dx, dy):
        """Draw the atom on the device context.

        Args:
            dc: The device context.
            style: The style to apply.
            x: The x-coordinate.
            y: The y-coordinate.
            dx: The width.
            dy: The height.

        Returns:
            bool: True if drawing was successful.
        """
        try:
            self.reg_id(dc)
            self.reg_end()
            dc2 = dc.subdc(x, y, self.width, self.height, True)
            for r in self.render_helpers:
                dc2 = r.render(dc2)
            dc2.play_str(self.draw_txt)
            return True
        except Exception as e:
            raise RuntimeError(f"Error drawing atom: {e}")


register_tag_map("vimg", VectorImg)



import django.forms.widgets


class ImgFileInput(django.forms.widgets.ClearableFileInput):
    """
    A custom file input widget for handling image file inputs.
    This widget extends ClearableFileInput to handle image files specifically.
    """

    def format_value(self, value):
        """
        Format the value to be displayed in the widget.

        Args:
            value: The value to be formatted.

        Returns:
            The formatted value.
        """
        return value

    def value_from_datadict(self, data, files, name):
        """
        Extract the value from the data dictionary.

        Args:
            data: The data dictionary containing form data.
            files: The files dictionary containing uploaded files.
            name: The name of the field.

        Returns:
            The value extracted from the data or files dictionary, or None if not found.
        """
        try:
            if name in data:
                return data[name]
            elif name in files:
                return files[name]
            else:
                return None
        except Exception as e:
            # Log the error or handle it as needed
            raise ValueError(f"Error extracting value from datadict: {e}")



"""Helper class for JSON encoding and decoding."""

import json
from urllib.parse import quote_plus, unquote_plus
import datetime
from decimal import Decimal


class ComplexEncoder(json.JSONEncoder):
    """Custom JSON encoder that handles complex types like datetime and Decimal."""

    standard_types = (
        "list",
        "unicode",
        "str",
        "int",
        "long",
        "float",
        "bool",
        "NoneType",
    )

    def default(self, obj):
        """Convert non-standard types to a JSON-compatible format."""
        if obj.__class__.__name__ not in self.standard_types:
            if isinstance(obj, datetime.datetime):
                return {"object": repr(obj).replace(", tzinfo=<UTC>", "")}
            elif hasattr(obj, "tolist"):  # Handle numpy arrays or similar
                return obj.tolist()
            else:
                return {"object": repr(obj)}
        return super().default(obj)


def as_complex(dct):
    """Convert JSON objects back to their original Python types."""
    if "object" in dct:
        try:
            return eval(dct["object"])
        except (NameError, SyntaxError):
            return None
    return dct


def dumps(obj):
    """Encode a Python object to a JSON string and URL-encode it.

    Args:
        obj: Python object to encode.

    Returns:
        str: URL-encoded JSON string.
    """
    try:
        return quote_plus(json.dumps(obj, cls=ComplexEncoder))
    except (TypeError, ValueError) as e:
        raise ValueError(f"Failed to encode object: {e}")


def loads(json_str):
    """Decode a URL-encoded JSON string to a Python object.

    Args:
        json_str: URL-encoded JSON string.

    Returns:
        object: Decoded Python object.
    """
    try:
        return json.loads(unquote_plus(json_str), object_hook=as_complex)
    except (ValueError, json.JSONDecodeError) as e:
        raise ValueError(f"Failed to decode JSON string: {e}")


def json_dumps(obj, indent=None):
    """Encode a Python object to a JSON string.

    Args:
        obj: Python object to encode.
        indent: Optional indentation for pretty-printing.

    Returns:
        str: JSON-encoded string.
    """
    try:
        return json.dumps(obj, cls=ComplexEncoder, indent=indent)
    except (TypeError, ValueError) as e:
        raise ValueError(f"Failed to encode object: {e}")


def json_loads(json_str):
    """Decode a JSON string to a Python object.

    Args:
        json_str: JSON-encoded string.

    Returns:
        object: Decoded Python object.
    """
    try:
        return json.loads(json_str, object_hook=as_complex)
    except (ValueError, json.JSONDecodeError) as e:
        raise ValueError(f"Failed to decode JSON string: {e}")


class ComplexDecoder(json.JSONDecoder):
    """Custom JSON decoder that uses the as_complex function for object conversion."""

    def decode(self, s):
        """Decode a JSON string to a Python object.

        Args:
            s: JSON-encoded string.

        Returns:
            object: Decoded Python object.
        """
        return json_loads(s)



from graphene import Node
from graphene_django.filter import DjangoFilterConnectionField
from graphene_django.types import DjangoObjectType


def add_graphql_to_class(model, filter_fields, query_class):
    """
    Adds GraphQL schema definitions to a Django model class.

    Args:
        model: The Django model class to which GraphQL schema will be added.
        filter_fields: Fields to be used for filtering in GraphQL queries.
        query_class: The GraphQL query class where the schema will be added.

    Raises:
        AttributeError: If the model does not have the required attributes.
    """

    _model = model
    _filter_fields = filter_fields

    if hasattr(_model._meta, "app_label"):
        app_label = getattr(_model._meta, "app_label")
    else:
        app_label = ""

    try:
        app_label = model._meta.app_label if hasattr(model._meta, "app_label") else ""

        class Meta:
            nonlocal _model, _filter_fields
            model = _model
            interfaces = (Node,)
            filter_fields = _filter_fields

        # Dynamically create a new DjangoObjectType class for the model
        ModelType = type(
            f"{app_label}__{model.__name__}__class",
            (DjangoObjectType,),
            {"Meta": Meta},
        )

        # Add the new type to the query class
        setattr(query_class, f"{app_label}__{model.__name__}", Node.Field(ModelType))
        setattr(
            query_class,
            f"{app_label}__{model.__name__}All",
            DjangoFilterConnectionField(ModelType),
        )

    except AttributeError as e:
        raise AttributeError(f"Model is missing required attributes: {e}")



from urllib.parse import urlencode
from typing import Any, Dict, List, Optional, Tuple, Union


class DictParm:
    """A class to handle dictionary parameters."""

    def __init__(self, data: Dict[str, Any]) -> None:
        """Initialize with a dictionary."""
        self.data = data

    def get_parm(self, param: str) -> Any:
        """Get a parameter from the dictionary."""
        if param not in self.data:
            raise KeyError(f"Parameter '{param}' not found in dictionary.")
        return self.data[param]

    def has_parm(self, param: str) -> bool:
        """Check if a parameter exists in the dictionary."""
        return param in self.data


def convert_param(param: Any) -> Union[str, List, bool]:
    """Convert a parameter to a suitable format for URL encoding."""
    if hasattr(param, "__class__") and param.__class__.__name__ == "DateTime":
        return str(param)[:10]
    if isinstance(param, (list, bool)):
        return param
    return str(param)


def dict_from_param(param: DictParm, fields: List[str]) -> Dict[str, Any]:
    """Create a dictionary from a list of fields using the given DictParm object."""
    return {field: param.get_parm(field) for field in fields if param.has_parm(field)}


def create_parm(
    address: str, dic: DictParm, no_encode: bool = False
) -> Optional[Tuple[str, str, Union[Dict, str]]]:
    """Create parameters from the address and dictionary."""
    parts = address.split("|")
    if len(parts) <= 1:
        return None

    params = parts[1].split(",")
    separator = "&" if "?" in address else "?"
    encoded_params = {}

    for param in params:
        if dic.has_parm(param):
            value = dic.get_parm(param)
            if value is not None:
                if "__" in param:
                    base_param = param.split("__")[0]
                    if base_param in encoded_params:
                        if isinstance(encoded_params[base_param], list):
                            encoded_params[base_param].append(convert_param(value))
                        else:
                            encoded_params[base_param] = [
                                encoded_params[base_param],
                                convert_param(value),
                            ]
                    else:
                        encoded_params[base_param] = convert_param(value)
                else:
                    encoded_params[param] = convert_param(value)

    if no_encode:
        return parts[0], separator, encoded_params
    else:
        return parts[0], separator, urlencode(encoded_params, doseq=True)


def create_post_param(address: str, dic: DictParm) -> Tuple[str, Dict[str, Any]]:
    """Create POST parameters from the address and dictionary."""
    parts = address.split("|")
    if len(parts) > 1:
        params = parts[1].split(",")
        return parts[0], dict_from_param(dic, params)
    return parts[0], {}



import io
import os
from decimal import Decimal
from datetime import date, datetime
import xlsxwriter
from pytigon_lib.schhtml.basedc import BaseDc, BaseDcInfo
from pytigon_lib.schfs.vfstools import get_temp_filename

IMAGE = None


class XlsxDc(BaseDc):
    def __init__(
        self,
        ctx=None,
        calc_only=False,
        width=8.5,
        height=11,
        output_name=None,
        output_stream=None,
        scale=1.0,
        notify_callback=None,
        record=False,
    ):
        super().__init__(
            calc_only,
            -1,
            -1,
            output_name,
            output_stream,
            scale,
            notify_callback,
            record,
        )
        self.dc_info = XlsxDcinfo(self)
        self.type = None
        self.width = width if width >= 0 else -1
        self.height = height if height >= 0 else 1000000000
        self.last_style_tab = None
        self.handle_html_directly = True
        self.temp_file_name = get_temp_filename()
        self.document = xlsxwriter.Workbook(self.temp_file_name)
        self.page_width = width
        self.page_height = height
        self.map_start_tag = {"body": self.body, "div": self.div}
        self.map_end_tag = {
            "tr": self.tr,
            "td": self.td,
            "th": self.th,
            "h1": self.h1,
            "h2": self.h2,
            "h3": self.h3,
            "h4": self.h4,
            "h5": self.h5,
            "h6": self.h6,
            "p": self.p,
            "img": self.image,
            "body": self.end_body,
        }
        self.last_ = None
        self.styles_cache = {}

        if self.notify_callback:
            self.notify_callback("start", {"dc": self})

    def close(self):
        if self.notify_callback:
            self.notify_callback("end", {"dc": self})
        self.document.close()
        try:
            with open(self.temp_file_name, "rb") as f_in:
                if self.output_stream:
                    self.output_stream.write(f_in.read())
                elif self.output_name:
                    with open(self.output_name, "wb") as f_out:
                        f_out.write(f_in.read())
        finally:
            os.unlink(self.temp_file_name)

    def annotate(self, what, data):
        element = data.get("element")
        if element and element.parent:
            if what == "start_tag" and element.tag in self.map_start_tag:
                self.map_start_tag[element.tag](element, element.parent)
            elif what == "end_tag" and element.tag in self.map_end_tag:
                self.map_end_tag[element.tag](element, element.parent)

    def _set_width(self, worksheet, col, value):
        width = int(
            value.replace("%", "")
            .replace("px", "")
            .replace("rem", "")
            .replace("em", "")
        )
        worksheet.set_column(col, col, width)

    def _set_height(self, worksheet, row, value):
        if value > 0:
            worksheet.set_row(row, value)

    def _get_color(self, color_str):
        if color_str:
            r, g, b = self.rgbfromhex(color_str)
            return f"#{r:02X}{g:02X}{b:02X}"
        return None

    def _get_style(self, element):
        style_str = self.dc_info.styles[element.style]
        border = int(element.attrs.get("border", 0))
        style_str += f";{border}"
        align = element.attrs.get("align", element.attrs.get("text-align", ""))
        style_str += f";{align}"
        bgcolor = element.attrs.get(
            "bgcolor",
            element.attrs.get("background-color", element.attrs.get("background", "")),
        )
        style_str += f";{bgcolor}"
        brcolor = element.attrs.get("border-color", "")
        style_str += f";{brcolor}"
        fmt = element.attrs.get("format", "")
        style_str += f";{fmt}"

        if style_str not in self.styles_cache:
            style_tab = style_str.split(";")
            fmt = self.document.add_format()
            if style_tab[3] == "1":
                fmt.set_italic()
            if style_tab[4] == "1":
                fmt.set_bold()
            fmt.set_font_name(style_tab[1])
            fmt.set_font_size(
                int(
                    (self.scale * (self.base_font_size * 72) * int(style_tab[2]))
                    / (96 * 100.0)
                )
            )
            fcolor = self._get_color(style_tab[0])
            if fcolor:
                fmt.set_font_color(fcolor)
            fmt.set_border(int(style_tab[6]))
            if style_tab[7] == "center":
                fmt.set_align("center")
            elif style_tab[7] == "right":
                fmt.set_align("right")
            else:
                fmt.set_align("left")
            bgcolor = self._get_color(style_tab[8])
            if bgcolor:
                fmt.set_bg_color(bgcolor)
            brcolor = self._get_color(style_tab[9])
            if brcolor:
                fmt.set_border_color(brcolor)
            if style_tab[10]:
                fmt.set_num_format(style_tab[10])
            self.styles_cache[style_str] = fmt
        return self.styles_cache[style_str]

    def _process_atom_list(self, element):
        return (
            "".join(
                atom.data
                for atom in element.atom_list.atom_list
                if isinstance(atom.data, str)
            )
            if element.atom_list and element.atom_list.atom_list
            else ""
        )

    def body(self, element, parent):
        if not hasattr(element, "worksheet"):
            title = element.attrs.get("title", "")
            element.worksheet = self.document.add_worksheet(title)
            element.status = [0, 0]
        if "cellwidth" in element.attrs:
            for col, value in enumerate(element.attrs["cellwidth"].split(";")):
                self._set_width(element.worksheet, col, value)

    def end_body(self, element, parent):
        if self.notify_callback:
            self.notify_callback(
                "worksheet",
                {"dc": self, "worksheet": element.worksheet, "status": element.status},
            )

    def div(self, element, parent):
        if parent.tag == "body":
            element.worksheet = parent.worksheet
            element.status = parent.status

    def tr(self, element, parent):
        parent.parent.status[0] += 1
        parent.parent.status[1] -= len(element.td_list)

    def td(self, element, parent):
        style = self._get_style(element)
        txt = self._process_atom_list(element)
        td_class = element.attrs.get("class", "")
        row, col = parent.parent.parent.status[0], parent.parent.parent.status[1]

        if td_class in ("int", "float", "decimal"):
            num = (
                int(txt)
                if td_class == "int"
                else float(txt)
                if td_class == "float"
                else Decimal(txt)
            )
            parent.parent.parent.worksheet.write_number(row, col, num, style)
        elif td_class in ("date", "datetime"):
            d = (
                date.fromisoformat(txt)
                if td_class == "date"
                else datetime.fromisoformat(txt)
            )
            parent.parent.parent.worksheet.write_datetime(row, col, d, style)
        elif td_class == "bool":
            b = txt.strip() not in ("", "0", "False", "None")
            parent.parent.parent.worksheet.write_boolean(row, col, b, style)
        elif td_class == "formula":
            parent.parent.parent.worksheet.write_formula(row, col, txt, style)
        elif td_class == "url":
            parent.parent.parent.worksheet.write_url(row, col, txt, style)
        else:
            parent.parent.parent.worksheet.write(row, col, txt, style)

        self._set_height(parent.parent.parent.worksheet, row, style.font_size + 4)
        parent.parent.parent.status[1] += 1
        parent.td_list.append(self)

    def th(self, element, parent):
        if "width" in element.attrs:
            self._set_width(
                parent.parent.parent.worksheet,
                parent.parent.parent.status[1],
                element.attrs["width"],
            )
        self.td(element, parent)

    def h(self, element, parent, level):
        if hasattr(parent, "worksheet"):
            style = self._get_style(element)
            txt = self._process_atom_list(element)
            parent.worksheet.write(parent.status[0], parent.status[1], txt, style)
            self._set_height(parent.worksheet, parent.status[0], style.font_size + 4)
            parent.status[0] += 1

    def p(self, element, parent):
        if parent.tag == "body":
            style = self._get_style(element)
            txt = self._process_atom_list(element)
            parent.worksheet.write(parent.status[0], parent.status[1], txt, style)
            self._set_height(parent.worksheet, parent.status[0], style.font_size + 4)
            parent.status[0] += 1

    def image(self, element, parent):
        if element.img and parent.tag in ("body", "div"):
            img_stream = io.BytesIO(element.img)
            parent.worksheet.insert_image(
                parent.status[0],
                parent.status[1],
                "https://pytigon.eu/pytigon.png",
                {"image_data": img_stream},
            )
            w, h = self.dc_info.get_img_size(element.img)
            self._set_height(parent.worksheet, parent.status[0], h)
            parent.status[0] += 1

    def h1(self, element, parent):
        self.h(element, parent, 0)

    def h2(self, element, parent):
        self.h(element, parent, 1)

    def h3(self, element, parent):
        self.h(element, parent, 2)

    def h4(self, element, parent):
        self.h(element, parent, 3)

    def h5(self, element, parent):
        self.h(element, parent, 4)

    def h6(self, element, parent):
        self.h(element, parent, 5)


class XlsxDcinfo(BaseDcInfo):
    def __init__(self, dc):
        super().__init__(dc)

    def get_text_height(self, word, style):
        return 1

    def get_img_size(self, png_data):
        global IMAGE
        if not IMAGE:
            from PIL import Image as IMAGE
        try:
            png_stream = io.BytesIO(png_data)
            image = IMAGE.open(png_stream)
            return image.size if image else (0, 0)
        except Exception:
            return (0, 0)



from .htmltools import superstrip
import re


def comment_remover(text):
    """Remove comments from the given text."""

    def replacer(match):
        s = match.group(0)
        return "" if s.startswith("/") else s

    pattern = re.compile(
        r'//.*?$|/\*.*?\*/|\'(?:\\.|[^\\\'])*\'|"(?:\\.|[^\\"])*"',
        re.DOTALL | re.MULTILINE,
    )
    return re.sub(pattern, replacer, text)


class CssPos:
    """Represents a CSS position with attributes and parent relationships."""

    def __init__(self, line, attrs):
        self.tag = superstrip(line[-1])
        self.parents = {}
        self.attrs = attrs if len(line) == 1 else {}
        if len(line) > 1:
            parent = CssPos(line[:-1], attrs)
            self.parents[parent.key()] = parent

    def key(self):
        """Return the key for this CSS position."""
        return self.tag

    def _extend_dict(self, dest, source):
        """Extend the destination dictionary with the source dictionary."""
        dest.update(source)

    def extend(self, line, attrs):
        """Extend the CSS position with additional attributes."""
        if line:
            key_main = line[-1]
            if key_main in self.parents:
                self.parents[key_main].extend(line[:-1], attrs)
            else:
                parent = CssPos(line, attrs)
                self.parents[parent.key()] = parent
        else:
            self.attrs.update(attrs)

    def _get_dict_from_parent(self, key, ret_attrs, obj):
        """Get attributes from a parent CSS position."""
        if key in self.parents and obj.get_parent():
            self._extend_dict(ret_attrs, self.parents[key].get_dict(obj.get_parent()))

    def get_dict(self, obj):
        """Get the combined attributes for the given object."""
        ret_attrs = self.attrs.copy()
        if obj:
            self._get_dict_from_parent(obj.get_tag(), ret_attrs, obj)
            if obj.get_cls():
                self._get_dict_from_parent("." + obj.get_cls(), ret_attrs, obj)
                self._get_dict_from_parent(
                    obj.get_tag() + "." + obj.get_cls(), ret_attrs, obj
                )
            if obj.get_id():
                self._get_dict_from_parent("#" + obj.get_id(), ret_attrs, obj)
                self._get_dict_from_parent(
                    obj.get_tag() + "." + obj.get_id(), ret_attrs, obj
                )
        return ret_attrs

    def test_print(self, indent):
        """Print the CSS position for testing purposes."""
        tab = indent * " "
        print(f"{tab}{self.key()}:")
        print(f"{tab}attrs:")
        for key, value in self.attrs.items():
            print(f"{tab}    {key}: {value}")
        print(f"{tab}parents:")
        for key, parent in self.parents.items():
            print(f"{tab}    key:")
            parent.test_print(indent + 8)


class Css:
    """Represents a CSS stylesheet."""

    def __init__(self):
        self.csspos_dict = {}
        self._act_dict = {}
        self._act_keys = []

    def _append_keys(self):
        """Append the current keys to the CSS position dictionary."""
        if self._act_keys:
            for pos in self._act_keys:
                lastkey = pos[-1]
                if lastkey in self.csspos_dict:
                    self.csspos_dict[lastkey].extend(pos[:-1], self._act_dict)
                else:
                    self.csspos_dict[lastkey] = CssPos(pos, self._act_dict)
        self._act_keys = []
        self._act_dict = {}

    def parse_indent_str(self, s):
        """Parse a string with indentation-based CSS."""
        for line in s.splitlines():
            if not line:
                continue
            indent = line[0] == " "
            line = superstrip(line.split("//")[0])
            if not line:
                continue
            if indent:
                key, *value = line.split(":")
                self._act_dict[key.strip()] = value[0].strip() if value else "0"
            else:
                if self._act_keys:
                    self._append_keys()
                self._act_keys = [
                    pos.strip().lower().split() for pos in line.split(",")
                ]
                self._act_dict = {}
        if self._act_keys:
            self._append_keys()

    def _strip_list(self, lst):
        """Strip whitespace from each element in the list."""
        return [item.strip() for item in lst]

    def _handle_section(self, section):
        """Handle a single CSS section."""
        selector, *content = section.split("{")
        if not content:
            return
        self._act_keys = [
            pos.strip().lower().split() for pos in superstrip(selector).split(",")
        ]
        self._act_dict = {}
        for prop in content[0].split(";"):
            key, *value = self._strip_list(prop.split(":"))
            if key:
                self._act_dict[key] = value[0] if value else "0"
        if self._act_keys:
            self._append_keys()

    def parse_str(self, s):
        """Parse a CSS string."""
        s2 = comment_remover(s)
        for section in superstrip(s2).split("}"):
            self._handle_section(section)

    def test_print(self):
        """Print the CSS for testing purposes."""
        tmp = CssPos([""], {})
        tmp.parents = self.csspos_dict
        tmp.test_print(0)

    def get_dict(self, obj):
        """Get the combined attributes for the given object."""
        tmp = CssPos([""], {})
        tmp.parents = self.csspos_dict
        return tmp.get_dict(obj)



def sch_import(name):
    """
    Dynamically import a module by its name.

    Args:
        name (str): The full module name to import (e.g., 'package.module').

    Returns:
        module: The imported module.

    Raises:
        ImportError: If the module or any of its components cannot be imported.
    """
    try:
        mod = __import__(name)
        components = name.split(".")
        for comp in components[1:]:
            mod = getattr(mod, comp)
        return mod
    except (ImportError, AttributeError) as e:
        raise ImportError(f"Failed to import module '{name}': {e}")



from django.http import HttpResponse, JsonResponse
from oauth2_provider.views import ProtectedResourceView
from oauth2_provider.views.mixins import ProtectedResourceMixin
from graphene_django.views import GraphQLView
import json


class OAuth2ProtectedResourceMixin(ProtectedResourceView):
    """
    Mixin to protect resources using OAuth2 authentication.
    Handles OPTIONS preflight requests and verifies user authentication.
    """

    def dispatch(self, request, *args, **kwargs):
        """
        Dispatch method to handle incoming requests.
        """
        try:
            # Allow preflight OPTIONS requests to pass through
            if request.method.upper() == "OPTIONS":
                return super(ProtectedResourceMixin, self).dispatch(
                    request, *args, **kwargs
                )

            # Verify if the request is valid and the user is authenticated
            if request.user.is_authenticated:
                valid = True
                user = request.user
            else:
                valid, r = self.verify_request(request)
                user = r.user

            if valid:
                request.resource_owner = user
                return super(ProtectedResourceMixin, self).dispatch(
                    request, *args, **kwargs
                )
            else:
                # Return authentication failure response
                message = {"evr-api": {"errors": ["Authentication failure"]}}
                return JsonResponse(message, status=401)

        except Exception as e:
            # Handle unexpected errors
            message = {"evr-api": {"errors": [str(e)]}}
            return JsonResponse(message, status=500)


class OAuth2ProtectedGraph(OAuth2ProtectedResourceMixin, GraphQLView):
    """
    View to protect GraphQL endpoints using OAuth2 authentication.
    """

    @classmethod
    def as_view(cls, *args, **kwargs):
        """
        Class method to create a view instance.
        """
        view = super(OAuth2ProtectedGraph, cls).as_view(*args, **kwargs)
        return view



import os
import sys
import zipfile
import shutil
import configparser
import multiprocessing
from pytigon_lib.schfs import extractall
from pytigon_lib.schtools.process import py_manage
from pytigon_lib.schtools.process import py_run
from pytigon_lib.schtools.nim_integration import install_nim


def _mkdir(path, ext=None):
    if ext:
        p = os.path.join(path, ext)
    else:
        p = path
    if not os.path.exists(p):
        try:
            os.mkdir(p)
        except:
            pass


def upgrade_test(zip_path, out_path):
    if os.path.exists(zip_path):
        archive = zipfile.ZipFile(zip_path, "r")
        cfg_txt = archive.read("install.ini").decode("utf-8")
        cfg = configparser.ConfigParser()
        cfg.read_string(cfg_txt)
        t1 = cfg["DEFAULT"]["GEN_TIME"]
        ini2 = os.path.join(out_path, "install.ini")
        if os.path.exists(ini2):
            cfg2 = configparser.ConfigParser()
            cfg2.read(ini2)
            t2 = cfg2["DEFAULT"]["GEN_TIME"]
            if t2 < t1:
                return True
        else:
            return False
    return False


def pip_install(pip_str, prjlib, confirm=False, upgrade=False):
    packages = [x.strip() for x in pip_str.split(" ") if x]
    print("pip install: ", pip_str)
    exit_code, output_tab, err_tab = py_run(
        [
            "-m",
            "pip",
            "--disable-pip-version-check",
            "install",
            f"--target={prjlib}",
        ]
        + (
            [
                "--upgrade",
            ]
            if upgrade
            else []
        )
        + packages
    )
    success = False
    if output_tab:
        for pos in output_tab:
            if pos:
                print("pip info: ", pos)
            if "Successfully installed" in pos:
                success = True
    if err_tab:
        for pos in err_tab:
            if pos:
                print("pip error: ", pos)

    if success and confirm:
        return True
    else:
        return False


def build_all(path):
    ret = True
    for root, dirs, files in os.walk(path):
        for name in files:
            if name.endswith("_build.py"):
                p = os.path.join(root, name)
                with open(p, "rt") as f:
                    l = locals()
                    buf = f.read()
                    exec(buf, globals(), l)
                    if "build" in l:
                        x = l["build"](path=p.replace("_build.py", ".nim"))
                        if not x:
                            ret = False
    return ret


def upgrade_local_libs():
    from django.conf import settings

    prjlib = os.path.join(settings.DATA_PATH, settings.PRJ_NAME, "prjlib")
    config_file = os.path.join(settings.PRJ_PATH, settings.PRJ_NAME, "install.ini")
    if os.path.exists(config_file):
        config = configparser.ConfigParser()
        config.read(config_file)
        if "DEFAULT" in config:
            pip_str = config["DEFAULT"].get("PIP", "")
            if pip_str:
                pip_install(pip_str, prjlib, confirm=True, upgrade=True)


SYS_COMMANDS = {
    "makeallmigrations",
    "migrate",
    "createautouser",
    "import_projects",
}


def init(prj, root_path, data_path, prj_path, static_app_path, paths=None):
    if prj == "_schall":
        return

    try:
        l = multiprocessing.Lock()
    except:
        l = None
    if l:
        l.acquire()

    _root_path = os.path.normpath(root_path)
    _data_path = os.path.normpath(data_path)
    _prj_path = os.path.normpath(prj_path)
    _static_app_path = os.path.normpath(static_app_path)
    # _base_compiler_path = os.path.join(_data_path, "ext_prg")
    # is_prj_path = True if os.path.exists(_prj_path) else False
    is_data_path = (
        True if os.path.exists(os.path.join(_data_path, "install.ini")) else False
    )
    is_dev = prj in ("schdevtools", "schmanage", "_schsetup")
    is_static_path = True if os.path.exists(_static_app_path) else False
    upgrade = False

    if is_data_path and is_dev:
        if upgrade_test(
            os.path.join(os.path.join(_root_path, "install"), ".pytigon.zip"),
            _data_path,
        ):
            upgrade = True
            print("Upgrade data")

    if not is_data_path:
        zip_file2 = os.path.join(os.path.join(_root_path, "install"), ".pytigon.zip")
        if not os.path.exists(_data_path):
            os.makedirs(_data_path)
        if os.path.exists(zip_file2) and is_dev:
            extractall(zipfile.ZipFile(zip_file2), _data_path)
        if not os.path.exists(os.path.join(_data_path, "media")):
            media_path = os.path.join(os.path.join(_data_path, "media"))
            os.makedirs(media_path)
            os.makedirs(os.path.join(media_path, "filer_public"))
            os.makedirs(os.path.join(media_path, "filer_private"))
            os.makedirs(os.path.join(media_path, "filer_public_tumbnails"))
            os.makedirs(os.path.join(media_path, "filer_private_thumbnails"))
        if not os.path.exists(os.path.join(_data_path, "doc")):
            doc_path = os.path.join(os.path.join(_data_path, "doc"))
            os.makedirs(doc_path)

        if SYS_COMMANDS.intersection(sys.argv):
            prjs = []
        elif is_dev:
            prjs = [ff for ff in os.listdir(_prj_path) if not ff.startswith("_")]
        else:
            prjs = [
                prj,
            ]

        tmp = os.getcwd()
        for app in prjs:
            path = os.path.join(_prj_path, app)
            if os.path.isdir(path):
                db_path = os.path.join(os.path.join(_data_path, app), f"{app}.db")
                os.chdir(path)
                print("python: pytigon: init: ", path)
                if not os.path.exists(db_path):
                    print("python: pytigon: init: create:", db_path)
                    exit_code, output_tab, err_tab = py_manage(
                        ["makeallmigrations"], False
                    )
                    if err_tab:
                        print(err_tab)
                    exit_code, output_tab, err_tab = py_manage(["migrate"], False)
                    if err_tab:
                        print(err_tab)
                    exit_code, output_tab, err_tab = py_manage(
                        ["createautouser"], False
                    )
                    if err_tab:
                        print(err_tab)
                    if app == "schdevtools":
                        print("python: pytigon: import_projects!")
                        exit_code, output_tab, err_tab = py_manage(
                            ["import_projects"], False
                        )
                        print("python: pytigon: projects imported!")
                        if err_tab:
                            print(err_tab)
        # install_nim(_data_path)
        os.chdir(tmp)

    if upgrade:
        zip_file2 = os.path.join(os.path.join(_root_path, "install"), ".pytigon.zip")
        if not os.path.exists(_data_path):
            os.makedirs(_data_path)
        if os.path.exists(zip_file2):
            extractall(zipfile.ZipFile(zip_file2), _data_path, exclude=[r".*\.db"])

    if not is_static_path:
        p2 = os.path.join(os.path.join(_root_path, "static"), "app")
        if os.path.exists(p2):
            shutil.copytree(p2, _static_app_path)

    _paths = [
        "",
        "cache",
        "plugins_cache",
        "_schall",
        "schdevtools",
        "prj",
        "temp",
        "static",
        prj,
    ]
    for p in _paths:
        _mkdir(_data_path, p)
    if paths:
        for p in paths:
            _mkdir(p)

    prjlib = os.path.join(_data_path, prj, "prjlib")
    if not os.path.exists(prjlib) or not os.path.exists(
        os.path.join(prjlib, "install.txt")
    ):
        ok = True
        if not os.path.exists(prjlib):
            os.mkdir(prjlib)
        config_file = os.path.join(prj_path, prj, "install.ini")
        if os.path.exists(config_file):
            config = configparser.ConfigParser()
            config.read(config_file)
            if "DEFAULT" in config:
                pip_str = config["DEFAULT"].get("PIP", "")
                if pip_str:
                    x = pip_install(pip_str, prjlib, confirm=True)
                    if not x:
                        ok = False
        x = build_all(os.path.join(_prj_path, prj))
        if not x:
            ok = False
        if ok:
            with open(os.path.join(prjlib, "install.txt"), "wt") as f:
                f.write("OK")

    if os.path.exists(prjlib):
        if prjlib not in sys.path:
            sys.path.append(prjlib)
        # if test1 or test2 or test3:
        #    ret = make(_data_path, os.path.join(_prj_path, prj), prj)
        #    if ret:
        #        for pos in ret:
        #            print(pos)
    syslib = os.path.join(_data_path, prj, "syslib")
    if not os.path.exists(syslib):
        os.makedirs(syslib)
        with open(os.path.join(syslib, "__init__.py"), "wt") as f:
            f.write(" ")
    if l:
        l.release()



import os
import sys
import django
from django.core.management import execute_from_command_line


def cmd(arg, from_main=False):
    """
    Execute a Django management command.

    Args:
        arg (str or list): The command to execute. If a string, it will be converted to a list.
        from_main (bool): If True, `arg` is treated as the full command line arguments.

    Raises:
        SystemExit: If the command execution fails.
    """
    try:
        if from_main:
            argv = arg
        else:
            argv = ["manage.py"] + ([arg] if isinstance(arg, str) else arg)

        os.environ.setdefault("DJANGO_SETTINGS_MODULE", "settings_app")
        execute_from_command_line(argv)
    except Exception as e:
        print(f"Error executing command: {e}", file=sys.stderr)
        sys.exit(1)


def syncdb():
    """Synchronize the database."""
    cmd("syncdb")


def help():
    """Display help information."""
    cmd("help")


if __name__ == "__main__":
    cmd(sys.argv, from_main=True)



import re
import shutil
import base64
from zipfile import ZipFile, ZIP_DEFLATED
from lxml import etree, html
from xml.sax.saxutils import escape
from pytigon_lib.schfs.vfstools import delete_from_zip

OFFICE_URN = "{urn:oasis:names:tc:opendocument:xmlns:office:1.0}"
TABLE_URN = "{urn:oasis:names:tc:opendocument:xmlns:table:1.0}"
TEXT_URN = "{urn:oasis:names:tc:opendocument:xmlns:text:1.0}"


def attr_get(attrs, key):
    """Get attribute value by key suffix."""
    for k in attrs.keys():
        if k.endswith(key):
            return attrs[k]
    return None


def transform_str(s):
    """Transform string by replacing special characters."""
    return s.replace("***", '"').replace("**", "'")


def inner_html(elem):
    return (elem.text or "") + "".join(
        [html.tostring(child).decode("utf-8") for child in elem.iterchildren()]
    )


class OdfDocTransform:
    """Transform ODF files."""

    def __init__(self, file_name_in, file_name_out=None):
        """Initialize OdfDocTransform with input and output file names."""
        self.file_name_in = file_name_in
        self.file_name_out = file_name_out or file_name_in.replace("_", "")
        self.process_tables = None
        self.doc_type = 1
        self.buf = None
        self.auto_cells = False

    def set_doc_type(self, doc_type):
        """Set document type: 0 - other, 1 - spreadsheet, 2 - writer."""
        self.doc_type = doc_type

    def set_process_tables(self, tables):
        """Set tables to process."""
        self.process_tables = tables

    def column_number(self):
        """Return template string for column increment."""
        return "{{ tbl.IncCol }}"

    def row_number(self, il=1):
        return """{{ tbl|args:%d|call:'IncRow' }}{{ tbl|args:1|call:'SetCol' }}""" % il

    def clear_row_col(self):
        return """{{ tbl|args:1|call:'SetRow' }}{{ tbl|args:1|call:'SetCol' }}"""

    def doc_process(self, doc, debug):
        """Process document type 2 (writer)."""
        pass

    def spreadsheet_process(self, doc, debug):
        """Process spreadsheet document."""
        self._process_annotations(doc, debug)
        self._process_table_cells(doc, debug)
        self._process_table_rows(doc)
        self._process_tables(doc)

    def _handle_annotation(self, comment_elem, comment_txt):
        """
        Handle annotations by processing comment text and inserting XML elements.

        This function processes the given comment text to determine the level of
        annotation and any positional offsets. It splits the comment text into
        individual values if necessary and inserts corresponding XML elements
        into the parent or grandparent nodes of the provided comment element.

        Args:
            comment_elem (etree.Element): The XML element representing the comment.
            comment_txt (str): The text content of the comment to be processed.

        The function supports annotations that begin with '^' or '$' to indicate
        hierarchical levels, and uses '@' to split the comment text into multiple
        parts for insertion at different positions.
        """

        parent = comment_elem.getparent()
        level = 0
        if comment_txt.startswith("."):
            if comment_elem is not None:
                if "table-cell" in comment_elem.tag:
                    for child in comment_elem.getchildren():
                        if "annotation" not in child.tag:
                            comment_elem.remove(child)
                    new_cell = etree.Element(TEXT_URN + "p")
                    new_cell.text = comment_txt[1:]
                    comment_elem.append(new_cell)
                else:
                    for child in comment_elem.getchildren():
                        if child.tag.endswith("v"):
                            comment_elem.remove(child)
                    comment_elem.append(
                        etree.XML("<is><t>%s</t></is>" % escape(comment_txt[1:]))
                    )
                    comment_elem.attrib["t"] = "inlineStr"
            return

        while comment_txt.startswith("^") or comment_txt.startswith("$"):
            level += 1
            offset = 0 if comment_txt.startswith("^") else 1
            comment_txt = comment_txt[1:]
        if "@" in comment_txt:
            x = comment_txt.split("@")
            values = [[x[0], level, 0], [x[1], level, 1]]
        else:
            values = [[comment_txt, level, 0]]

        for v, level, offset in values:
            if level < 1:
                parent.insert(
                    parent.index(comment_elem) + offset,
                    etree.XML("<tmp>%s</tmp>" % escape(v)),
                )
            else:
                gparent = parent
                while level > 0:
                    p = gparent
                    gparent = p.getparent()
                    level -= 1
                gparent.insert(
                    gparent.index(p) + offset,
                    etree.XML("<tmp>%s</tmp>" % escape(v)),
                )

    def _process_annotations(self, doc, debug):
        """Process annotations in the document."""
        for e in doc.findall(".//{*}annotation/{*}p"):
            element = e.getparent()
            data = e.text
            if not data:
                data = ""
                for txt in e.itertext():
                    if txt:
                        data += txt
            data = data.strip()
            if data and (
                data.startswith("^") or data.startswith("$") or data.startswith(".")
            ):
                self._handle_annotation(element.getparent(), data)
            parent = element.getparent()
            parent.remove(element)

    def _process_table_cells(self, doc, debug):
        """Process table cells in the document."""
        for element in doc.findall(".//{*}table-cell"):
            self._handle_repeated_columns(element)
            if attr_get(element.attrib, "value-type") == "string":
                self._handle_string_cell(element, debug)

    def _handle_repeated_columns(self, element):
        """Handle repeated columns in table cells."""
        nr = attr_get(element.attrib, "number-columns-repeated")
        if nr and int(nr) > 1000:
            element.set(TABLE_URN + "number-columns-repeated", "1000")

    def _handle_string_cell(self, element, debug):
        """Handle string type table cells."""
        txt = (
            etree.tostring(element, method="text", encoding="utf-8")
            .decode("utf-8")
            .strip()
        )
        if any(item in txt for item in (":=", ":*", ":?", "{{", "}}", "{%", "%}")):
            txt = transform_str(txt)
            if txt.startswith(":="):
                self._create_formula_cell(element, txt, debug)
            elif txt.startswith(":?"):
                self.auto_cells = True
                self._create_auto_cell(element, txt, debug)
            else:
                self._create_value_cell(element, txt, debug)

    def _create_formula_cell(self, element, txt, debug):
        """Create a new cell with a formula."""
        new_cell = etree.Element(TABLE_URN + "table-cell")
        new_cell.set(OFFICE_URN + "value-type", "float")
        new_cell.set(OFFICE_URN + "value", "0")
        new_cell.set(TABLE_URN + "formula", "of:=" + txt[2:])
        new_text = etree.Element(OFFICE_URN + "p")
        new_cell.append(new_text)
        if debug:
            self._add_annotation(new_cell, txt[2:].replace("^", ""))
        self._set_cell_style(element, new_cell)
        self._replace_cell(element, new_cell)

    def _create_value_cell(self, element, txt, debug):
        """Create a new cell with a value."""
        new_cell = etree.Element(TABLE_URN + "table-cell")
        if txt.startswith(":0"):
            new_cell.set(OFFICE_URN + "value-type", "float")
            new_cell.set(OFFICE_URN + "value", str(txt[2:]))
            new_text = etree.Element(TEXT_URN + "p")
            new_text.text = str(txt[2:])
            new_cell.append(new_text)
        else:
            new_cell.set(OFFICE_URN + "value-type", "string")
            new_text = etree.Element(TEXT_URN + "p")
            new_text.text = txt[2:] if txt.startswith(":*") else txt
            new_cell.append(new_text)
        if debug:
            self._add_annotation(new_cell, txt[2:] if txt.startswith(":*") else txt)
        self._set_cell_style(element, new_cell)
        self._replace_cell(element, new_cell)

    def _create_auto_cell(self, element, txt, debug):
        """Create a new cell with a value."""
        new_cell = etree.Element(TABLE_URN + "table-cell")
        new_cell.set(OFFICE_URN + "value-type", "string")
        new_text = etree.Element(TEXT_URN + "vauto")
        new_text.text = txt[2:] if txt.startswith(":?") else txt
        new_cell.append(new_text)
        if debug:
            self._add_annotation(new_cell, txt[2:] if txt.startswith(":?") else txt)
        self._set_cell_style(element, new_cell)
        self._replace_cell(element, new_cell)

    def _add_annotation(self, cell, text):
        """Add annotation to a cell."""
        new_annotate = etree.Element(OFFICE_URN + "annotation")
        new_text_a = etree.Element(TEXT_URN + "p")
        new_text_a.text = text
        new_annotate.append(new_text_a)
        cell.append(new_annotate)

    def _set_cell_style(self, element, new_cell):
        """Set style for a new cell."""
        style_name = attr_get(element.attrib, "style-name")
        if style_name:
            new_cell.set(TABLE_URN + "style-name", style_name)

    def _replace_cell(self, element, new_cell):
        """Replace old cell with a new one."""
        new_cell2 = etree.Element("tmp")
        new_cell2.append(new_cell)
        new_cell2.text = self.column_number()
        parent = element.getparent()
        parent[parent.index(element)] = new_cell2

    def _process_table_rows(self, doc):
        """Process table rows in the document."""
        for element in doc.findall(".//{*}table-row"):
            self._handle_repeated_rows(element)

    def _handle_repeated_rows(self, element):
        """Handle repeated rows in table."""
        nr = attr_get(element.attrib, "number-rows-repeated")
        nr = int(nr) if nr else 1
        if nr > 1000:
            element.set(TABLE_URN + "number-rows-repeated", "1000")
        new_cell = etree.Element("tmp")
        parent = element.getparent()
        parent[parent.index(element)] = new_cell
        new_cell.append(element)
        new_cell.text = self.row_number(nr)

    def _process_tables(self, doc):
        """Process tables in the document."""
        for element in doc.findall(".//{*}table"):
            self._reset_row_col(element)
            if (
                self.process_tables
                and attr_get(element.attrib, "name") not in self.process_tables
            ):
                self._remove_table(element)

    def _reset_row_col(self, element):
        """Reset row and column for a table."""
        new_cell = etree.Element("tmp")
        new_cell.text = self.clear_row_col()
        parent = element.getparent()
        parent[parent.index(element)] = new_cell
        new_cell.append(element)

    def _remove_table(self, element):
        """Remove a table from the document."""
        new_cell = etree.Element("tmp")
        parent = element.getparent()
        parent[parent.index(element)] = new_cell

    def process_template(self, doc_str, context):
        """Process template with context."""
        pass

    def extended_transformation(self, xml_name, script):
        """Apply extended transformation using a script."""
        xml = etree.fromstring(self.buf.encode("utf-8"))
        script(self, xml)
        self.buf = etree.tostring(xml, encoding="utf-8", xml_declaration=True).decode(
            "utf-8"
        )

    def repair_xml(self, sheet):
        auto_list = sheet.findall(".//{*}vauto")
        for pos in auto_list:
            parent = pos.getparent()
            txt = pos.text
            parent.remove(pos)
            if txt != "" and txt != None:
                if len(txt) == 10 and txt[4] == "-" and txt[7] == "-":
                    try:
                        new_text = etree.Element(TEXT_URN + "p")
                        parent.append(new_text)
                        parent.set(OFFICE_URN + "value-type", "date")
                        parent.set(OFFICE_URN + "date-value", txt[:10])
                        continue
                    except:
                        pass
                try:
                    x = float(txt)
                    new_text = etree.Element(TEXT_URN + "p")
                    parent.append(new_text)
                    parent.set(OFFICE_URN + "value-type", "float")
                    parent.set(OFFICE_URN + "value", txt)
                    continue
                except:
                    pass
                new_text = etree.Element(TEXT_URN + "p")
                new_text.text = escape(txt)
                parent.append(new_text)

    def process(self, context, debug):
        """Transform input file using context and debug mode."""
        try:
            shutil.copyfile(self.file_name_in, self.file_name_out)
            with ZipFile(self.file_name_out, "r") as z:
                doc_content = z.read("content.xml").decode("utf-8")

            if delete_from_zip(self.file_name_out, ["content.xml"]) == 0:
                return

            doc = etree.fromstring(
                doc_content.replace("&apos;", "'")
                .replace("_start_", "{{")
                .replace("_end_", "}}")
                .encode("utf-8")
            )

            if self.doc_type == 1:
                self.spreadsheet_process(doc, debug)
            elif self.doc_type == 2:
                self.doc_process(doc, debug)

            doc_str = (
                etree.tostring(doc, encoding="utf-8", xml_declaration=True)
                .decode("utf-8")
                .replace("<tmp>", "")
                .replace("</tmp>", "")
            )

            doc_str = re.sub(r"\^(.*?\(.*?\))", r"${\1}", doc_str)

            if "expr_escape" in context:
                doc_str = doc_str.replace("{{", "{% expr_escape ").replace("}}", " %}")

            x = self.process_template(doc_str, context) or doc_str

            if self.auto_cells:
                root = etree.XML(x.encode("utf-8"))
                self.repair_xml(root)
                x = etree.tostring(root, encoding="utf-8").decode("utf-8")

            files = []
            if "[[[" in x and "]]]" in x:
                data = [pos.split("]]]")[0] for pos in x.split("[[[")[1:]]
                data2 = [pos.split("]]]")[-1] for pos in x.split("[[[")]
                fdata = []
                for i, pos in enumerate(data, 1):
                    x = pos.split(",", 1)
                    ext = x[0].split(";")[0].split("/")[-1]
                    name = f"Pictures/pytigon_{i}.{ext}"
                    fdata.append(name)
                    files.append([name, x, ext])
                data3 = [None] * (len(data) + len(data2))
                data3[::2] = data2
                data3[1::2] = fdata
                x = "".join(data3)

            self.buf = x

            if "extended_transformations" in context:
                for pos in context["extended_transformations"]:
                    self.extended_transformation(pos[0], pos[1])

            with ZipFile(self.file_name_out, "a", ZIP_DEFLATED) as z:
                z.writestr("content.xml", self.buf.encode("utf-8"))
                for pos in files:
                    z.writestr(pos[0], base64.b64decode(pos[1].encode("utf-8")))

            return 1
        except Exception as e:
            print(f"Error processing file: {e}")
            return 0


if __name__ == "__main__":
    x = OdfDocTransform("./test.ods", "./test_out.ods")
    context = {"test": 1, "object_list": ["x1", "x2", "x3"]}
    x.process(context, False)



from pytigon_lib.schtools import schjson

# Command constants
CMD_INFO = 1
CMD_PAGE = 2
CMD_COUNT = 3
CMD_SYNC = 4
CMD_AUTO = 5
CMD_RECASSTR = 6
CMD_EXEC = 7


class Table:
    """Base class for server table interface."""

    def __init__(self):
        """Initialize table with default columns and types."""
        self.auto_cols = []
        self.col_length = [0]
        self.col_names = ["ID"]
        self.col_types = ["int"]
        self.default_rec = [0]

    def _info(self):
        """Return table metadata as JSON."""
        return schjson.dumps(
            {
                "auto_cols": self.auto_cols,
                "col_length": self.col_length,
                "col_names": self.col_names,
                "col_types": self.col_types,
                "default_rec": self.default_rec,
            }
        )

    def _page(self, nr, sort=None, value=None):
        """Return a page of records as JSON."""
        return schjson.dumps({"page": self.page(nr, sort, value)})

    def _rec_as_str(self, nr):
        """Return a record as a string in JSON format."""
        return schjson.dumps({"recasstr": self.rec_as_str(nr)})

    def _count(self, value=None):
        """Return the count of records as JSON."""
        return schjson.dumps({"count": self.count(value)})

    def _sync(self, update, insert, delete):
        """Synchronize records with the server."""
        try:
            for rec in update:
                self.update_rec(rec)
            for nr in delete:
                self.delete_rec(nr)
            for rec in insert:
                self.insert_rec(rec)
            return "OK"
        except Exception as e:
            return schjson.dumps({"error": str(e)})

    def _auto(self, col_name, col_names, rec):
        """Handle auto-column logic and return the result as JSON."""
        return schjson.dumps({"rec": self.auto(col_name, col_names, rec)})

    def _exec(self, value=None):
        """Execute a command and return the result."""
        try:
            ret = self.exec_command(value)
            return schjson.dumps(ret) if isinstance(ret, dict) else ret
        except Exception as e:
            return schjson.dumps({"error": str(e)})

    def page(self, nr, sort=None, value=None):
        """Return a page of records."""
        raise NotImplementedError

    def count(self, value=None):
        """Return the count of records."""
        raise NotImplementedError

    def rec_as_str(self, nr):
        """Return a record as a string."""
        raise NotImplementedError

    def insert_rec(self, rec):
        """Insert a record."""
        raise NotImplementedError

    def update_rec(self, rec):
        """Update a record."""
        raise NotImplementedError

    def delete_rec(self, nr):
        """Delete a record."""
        raise NotImplementedError

    def auto(self, col_name, col_names, rec):
        """Handle auto-column logic."""
        raise NotImplementedError

    def exec_command(self, value):
        """Execute a command."""
        raise NotImplementedError

    def command(self, cmd_dict):
        """Handle a command based on the command dictionary."""
        cmd = cmd_dict.get("cmd", CMD_PAGE)
        value = cmd_dict.get("value")
        nr = cmd_dict.get("nr", 0)
        sort = cmd_dict.get("sort")

        if cmd == CMD_INFO:
            return self._info()
        elif cmd == CMD_PAGE:
            return self._page(nr, sort, value)
        elif cmd == CMD_COUNT:
            return self._count(value)
        elif cmd == CMD_SYNC:
            return self._sync(
                schjson.loads(cmd_dict["update"]),
                schjson.loads(cmd_dict["insert"]),
                schjson.loads(cmd_dict["delete"]),
            )
        elif cmd == CMD_AUTO:
            return self._auto(
                cmd_dict["col_name"], cmd_dict["col_names"], cmd_dict["rec"]
            )
        elif cmd == CMD_RECASSTR:
            return self._rec_as_str(int(cmd_dict["nr"]))
        elif cmd == CMD_EXEC:
            return self._exec(value)
        else:
            return None


def str_cmp(x, y, ts):
    """Compare two strings based on the given sorting criteria."""
    (id, s) = ts[0]
    if x[id] > y[id]:
        return s
    if x[id] < y[id]:
        return -s
    if len(ts) > 1:
        return str_cmp(x, y, ts[1:])
    return 0


class TablePy(Table):
    """Python implementation of the Table class."""

    def __init__(self, table, col_names, col_typ, col_length, default_rec):
        """Initialize the table with data and metadata."""
        self.tab = table
        self.auto_cols = []
        self.col_length = [0] + col_length
        self.col_names = ["ID"] + col_names
        self.col_types = ["int"] + col_typ
        self.default_rec = [0] + default_rec

    def page(self, nr, sort=None, value=None):
        """Return a page of records."""
        tab = []
        tab2 = self.tab[nr * 256 : (nr + 1) * 256]
        for i, rec in enumerate(tab2):
            tab.append([nr * 256 + i] + rec)

        if sort:
            s = sort.split(",")
            ts = []
            for pos in s:
                if pos:
                    id = self.col_names.index(pos[1:] if pos[0] == "-" else pos)
                    ss = -1 if pos[0] == "-" else 1
                    ts.append((id, ss))
            tab.sort(key=lambda x: str_cmp(x, x, ts))

        return tab

    def count(self, value=None):
        """Return the count of records."""
        return len(self.tab)

    def insert_rec(self, rec):
        """Insert a record."""
        self.tab.append(rec[1:])

    def update_rec(self, rec):
        """Update a record."""
        self.tab[rec[0]] = rec[1:]

    def delete_rec(self, nr):
        """Delete a record."""
        self.tab.pop(nr)

    def auto(self, col_name, col_names, rec):
        """Handle auto-column logic."""
        pass



"""Module contains views for form processing."""

from django.http import HttpResponse
from django.template import RequestContext, loader
from django.conf import settings
from pytigon_lib.schviews.viewtools import render_to_response
from .perms import make_perms_url_test_fun
from .viewtools import render_to_response_ext


def form(
    request,
    app_name,
    form_class,
    template_name,
    object_id=None,
    form_end=False,
    param=None,
    mimetype=None,
):
    """Create and process a form view.

    Args:
        request: The HTTP request object.
        app_name: The name of the application.
        form_class: The form class to instantiate.
        template_name: The template to render.
        object_id: Optional object ID for the form.
        form_end: Optional flag indicating form end.
        param: Optional parameters for form processing.
        mimetype: Optional mimetype for the response.

    Returns:
        HttpResponse: The rendered response.
    """
    try:
        form_instance = None
        if hasattr(form_class, "get_form_arguments"):
            form_args = form_class.get_form_arguments(request)
            if form_args:
                form_instance = form_class(**form_args)

        if not form_instance:
            form_instance = form_class(request.POST or None, request.FILES or None)

        if hasattr(form_instance, "preprocess_request"):
            post_data = form_instance.preprocess_request(request)
        else:
            post_data = request.POST

        if post_data:
            if hasattr(form_instance, "init"):
                form_instance.init(request)

            if form_instance.is_valid():
                result = (
                    form_instance.process(request, param)
                    if param
                    else form_instance.process(request)
                )
                if not isinstance(result, dict):
                    return result

                result.update({"form": form_instance})
                if object_id:
                    result.update({"object_id": object_id})

                if hasattr(form_instance, "render_to_response"):
                    return form_instance.render_to_response(
                        request, template_name, RequestContext(request, result)
                    )
                else:
                    doc_type = result.get("doc_type", "html")
                    return render_to_response_ext(
                        request, template_name, context=result, doc_type=doc_type
                    )
            else:
                if hasattr(form_instance, "process_invalid"):
                    result = (
                        form_instance.process(request, param)
                        if param
                        else form_instance.process(request)
                    )
                    result.update({"form": form_instance})
                    if object_id:
                        result.update({"object_id": object_id})
                    return render_to_response(
                        template_name, context=result, request=request
                    )
                else:
                    return render_to_response(
                        template_name, context={"form": form_instance}, request=request
                    )
        else:
            if hasattr(form_instance, "init"):
                form_instance.init(request)
            if object_id:
                form_instance.object_id = object_id

            if hasattr(form_instance, "process_empty"):
                result = (
                    form_instance.process_empty(request, param)
                    if param
                    else form_instance.process_empty(request)
                )
                result["form"] = form_instance
            else:
                result = {"form": form_instance}
                if object_id:
                    result.update({"object_id": object_id})
                if param:
                    result.update(param)

            return render_to_response(template_name, context=result, request=request)
    except Exception as e:
        return HttpResponse(f"An error occurred: {str(e)}", status=500)


def form_with_perms(app):
    """Create a form view with permissions."""
    return make_perms_url_test_fun(app, form)


def list_and_form(
    request,
    queryset,
    form_class,
    template_name,
    table_always=True,
    paginate_by=None,
    page=None,
    allow_empty=True,
    extra_context=None,
    context_processors=None,
    template_object_name="obj",
    mimetype=None,
    param=None,
):
    """List and process a form view.

    Args:
        request: The HTTP request object.
        queryset: The queryset to display.
        form_class: The form class to instantiate.
        template_name: The template to render.
        table_always: Whether to always display the table.
        paginate_by: Number of items per page.
        page: The current page number.
        allow_empty: Whether to allow empty querysets.
        extra_context: Additional context data.
        context_processors: Context processors to apply.
        template_object_name: The name of the object in the template.
        mimetype: Optional mimetype for the response.
        param: Optional parameters for form processing.

    Returns:
        HttpResponse: The rendered response.
    """
    try:
        form_instance = form_class(request.POST or None)
        if request.POST and form_instance.is_valid():
            queryset = (
                form_instance.Process(request, queryset, param)
                if param
                else form_instance.Process(request, queryset)
            )
            extra_context = extra_context or {}
            extra_context.update({"form": form_instance})
        elif table_always:
            extra_context = extra_context or {}
            extra_context.update({"form": form_instance})
            if hasattr(form_instance, "ProcessEmpty"):
                queryset = (
                    form_instance.ProcessEmpty(request, queryset, param)
                    if param
                    else form_instance.ProcessEmpty(request, queryset)
                )

        return render_to_response(
            template_name,
            context={
                "form": form_instance,
                "object_list": queryset,
                **(extra_context or {}),
            },
            request=request,
        )
    except Exception as e:
        return HttpResponse(f"An error occurred: {str(e)}", status=500)


def direct_to_template(request, template, extra_context=None, mimetype=None, **kwargs):
    """Render a template directly with additional context.

    Args:
        request: The HTTP request object.
        template: The template to render.
        extra_context: Additional context data.
        mimetype: Optional mimetype for the response.
        **kwargs: Additional URL parameters.

    Returns:
        HttpResponse: The rendered response.
    """
    try:
        context = {"params": kwargs}
        if extra_context:
            context.update(
                {k: v() if callable(v) else v for k, v in extra_context.items()}
            )
        return render_to_response(template, context=context, request=request)
    except Exception as e:
        return HttpResponse(f"An error occurred: {str(e)}", status=500)
